# Copyright 2018 Recruit Communications Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class CompiledConstraint:
    """Compiled constraint.
    
    Args:
        polynomial (dict[:class:`BinaryProd`, float/:class:`Coefficient`]):
            Polynomial representation of constraint.
    """

    def __init__(self, polynomial):
        self.polynomial = polynomial

    def energy(self, binary_solution, feed_dict):
        """Returns the energy of constraint given a solution.
        
        Args:
            binary_solution (dict[label, bit]):
                Binary solution.

            feed_dict (dict[str, float]):
                The value of placeholders.
        
        Returns:
            float: Energy of constraint.
        """
        energy = 0.0
        for binary_prod, coeff in self.polynomial.items():
            evaluated_value = self._eval_if_not_float(coeff, feed_dict)
            energy += binary_prod.calc_product(binary_solution) * evaluated_value
        return energy

    @staticmethod
    def _eval_if_not_float(v, feed_dict):
        """If v is not float (i.e. v is :class:`Express`), returns an evaluated value.

        Args:
            v (float/:class:`Coefficient`):
                The value to be evaluated.

            feed_dict (dict[str, float]):
                The value of placeholders.

        Returns:
            float: Evaluated value of the input :obj:`v`:
        """
        if isinstance(v, float):
            return v
        else:
            return v.evaluate(feed_dict)



# Copyright 2018 Recruit Communications Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class Coefficient:
    """The value of QUBO as a function of :class:`Placeholder`.
    
    Energy of QUBO is defined as :math:`E(\\mathbf{x}) = \\sum_{ij} a_{ij}x_{i}x_{j}`.
    
    If the expression contains :class:`Placeholder`,
    you need to specify the value of placeholders to get the final QUBO.
    Each coefficient :math:`a_{ij}` of compiled QUBO is defined as :class:`Coefficient`.
    If you want to get the final value of :math:`a_{ij}`, you need to evaluate it with `feed_dict`.
    
    Args:
        terms (dict[:class:`PlaceholderProd`, float]): polynomial function of :class:`Placeholder`.
            The labels in :class:`PlaceholderProd` corresponds to labels of :class:`Placeholder`.
    
    Example:
        
        For example, a polynomial :math:`2ab+2` is represented as
        
        >>> from pyqubo import Coefficient, PlaceholderProd
        >>> coeff = Coefficient({PlaceholderProd({'a': 1, 'b': 1}): 2.0, PlaceholderProd({}): 2.0})
        
        If we specify the value of Placeholder as :math:`a=2, b=3`,
        then the evaluated value will be 14.0.
        
        >>> coeff.evaluate(feed_dict={'a': 2, 'b': 3})
        14.0
        
    """

    def __init__(self, terms):
        self.terms = terms

    def evaluate(self, feed_dict):
        """Returns evaluated value with `feed_dict`.
        
        Args:
            feed_dict (dict[str, float]): The value of placeholder.
        
        Returns:
            float
        """
        if not feed_dict:
            raise ValueError("No feed_dict are given. Specify the feed_dict.")
        result = 0.0
        for term, value in self.terms.items():
            prod = value
            for key, p in term.keys.items():
                if key in feed_dict:
                    prod *= feed_dict[key] ** p
                else:
                    raise ValueError("{key} is not specified in feed_dict. "
                                     "Set the value of {key}".format(key=key))
            result += prod
        return result



# Copyright 2018 Recruit Communications Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from operator import or_
import dimod
from six.moves import reduce


class CompiledQubo:
    """Compiled QUBO.

    Args:
        qubo (dict[label, :class:`Coefficient`/float]): QUBO
        offset (:class:`Coefficient`/float): Offset of QUBO

    Attributes:
        qubo (dict[label, :class:`Coefficient`/float]): QUBO
        offset (:class:`Coefficient`/float): Offset of QUBO

    This contains QUBO and the offset, but the value of the QUBO
    has not been evaluated yet. To get the final QUBO, you need to
    evaluate this QUBO by calling :func:`CompiledQubo.eval`.
    """

    def __init__(self, qubo, offset):
        self.qubo = qubo
        self.offset = offset

    @property
    def variables(self):
        """Unique labels contained in keys of QUBO."""
        return reduce(or_, [{i, j} for (i, j) in self.qubo.keys()])

    def evaluate(self, feed_dict):
        """Returns QUBO where the values are evaluated with feed_dict.

        Args:
            feed_dict (dict[str, float]): The value of :class:`Placeholder`.

        Returns:
            :class:`BinaryQuadraticModel`
        """
        evaluated_qubo = {}
        for k, v in self.qubo.items():
            evaluated_qubo[k] = CompiledQubo._eval_if_not_float(v, feed_dict)

        evaluated_offset = CompiledQubo._eval_if_not_float(self.offset, feed_dict)

        return dimod.BinaryQuadraticModel.from_qubo(
            evaluated_qubo, evaluated_offset)

    def __repr__(self):
        from pprint import pformat
        return "CompiledQubo({}, offset={})".format(pformat(self.qubo), self.offset)

    @staticmethod
    def _eval_if_not_float(v, feed_dict):
        """ If v is not float (i.e. v is :class:`Express`), returns an evaluated value.

        Args:
            v (float/:class:`Coefficient`):
                The value to be evaluated.
            feed_dict (dict[str, float]):
                The value of :class:`Placeholder`.

        Returns:
            float: Evaluated value of the input :obj:`v`:
        """
        if isinstance(v, float):
            return v
        else:
            return v.evaluate(feed_dict)



# Copyright 2018 Recruit Communications Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from operator import or_
import numpy as np
import dimod
from six.moves import reduce
from dimod.decorators import vartype_argument


class Model:
    """Model represents binary quadratic optimization problem.
    
    By compiling :class:`Express` object, you get a :class:`Model` object.
    It contains the information about QUBO (or equivalent Ising Model),
    and it also has the function to decode the solution
    into the original variable structure.
    
    Note:
        We do not need to create this object directly. Instead,
        we get this by compiling `Express` objects.
    
    Args:
        compiled_qubo (:class:`.CompiledQubo`):
            Compiled QUBO. If we want to get the final QUBO, we need to evaluate this QUBO
            by specifying the value of placeholders. See :func:`CompiledQubo.eval()`.
            
        structure (`dict[label, Tuple(key1, key2, key3, ...)]`):
            It defines the mapping of the variable used in :func:`decode_solution`.
            A solution of `label` is mapped to
            :obj:`decoded_solution[key1][key2][key3][...]`.
            For more details, see :func:`decode_solution()`.
        
        constraints (dict[label, polynomial_term]):
            It contains constraints of the problem. `label` is each constraint name and
            `polynomial_term` is corresponding polynomial which should be zero when the
            constraint is satisfied.
        
    Attributes:
        variable_order (list):
            The list of labels. The order is corresponds to the index of QUBO or Ising model.
        
        index2label (dict[int, label]):
            The dictionary which maps an index to a label.
        
        label2index (dict[label, index]):
            The dictionary which maps a label to an index.
    """

    def __init__(self, compiled_qubo, structure, constraints):
        self.compiled_qubo = compiled_qubo
        self.structure = structure
        self.constraints = constraints
        self.variable_order = sorted(self.compiled_qubo.variables)
        self.index2label = dict(enumerate(self.variable_order))
        self.label2index = {v: k for k, v in self.index2label.items()}

    def __repr__(self):
        from pprint import pformat
        return "Model({}, structure={})".format(repr(self.compiled_qubo), str(pformat(self.structure)))

    @vartype_argument('vartype')
    def _parse_solution(self, solution, vartype):
        """Parse solution.
        
        Args:
            solution (list[bit]/dict[label, bit]/dict[index, bit]):
                The solution returned from solvers.
            
            vartype (:class:`dimod.Vartype`/str/set, optional):
                Variable type of the solution. Accepted input values
                    * :class:`.Vartype.SPIN`, ``'SPIN'``, ``{-1, 1}``
                    * :class:`.Vartype.BINARY`, ``'BINARY'``, ``{0, 1}``
        
        Returns:
             dict[label, bit]: dictionary of label and binary bit.
        """

        if isinstance(solution, list) or isinstance(solution, np.ndarray):
            if len(self.variable_order) != len(solution):
                raise ValueError("Illegal solution. Length of the solution is different from"
                                 "that of self.variable_order.")
            dict_binary_solution = dict(zip(self.variable_order, solution))
        elif isinstance(solution, dict):

            if set(solution.keys()) == set(self.variable_order):
                dict_binary_solution = solution

            elif set(solution.keys()) == set(range(len(self.variable_order))):
                dict_binary_solution = {self.index2label[index]: v for index, v in solution.items()}

            else:
                raise ValueError("Illegal solution. The keys of the solution"
                                 " should be same as self.variable_order")
        else:
            raise TypeError("Unexpected type of solution.")

        if vartype == dimod.SPIN:
            dict_binary_solution = {k: (v + 1) / 2 for k, v in dict_binary_solution.items()}

        return dict_binary_solution

    @vartype_argument('vartype')
    def energy(self, solution, vartype, feed_dict=None):
        """Returns energy of the solution.
        
        Args:
            solution (list[bit]/dict[label, bit]/dict[index, bit]):
                The solution returned from solvers.
            
            vartype (:class:`dimod.Vartype`/str/set, optional):
                Variable type of the solution.
                Accepted input values:
                
                * :class:`.Vartype.SPIN`, ``'SPIN'``, ``{-1, 1}``
                * :class:`.Vartype.BINARY`, ``'BINARY'``, ``{0, 1}``
            
            feed_dict (dict[str, float]):
                Specify the placeholder values.
        
        Returns:
            float: energy of the solution. 
        """
        dict_binary_solution = self._parse_solution(solution, vartype)
        bqm = self.compiled_qubo.evaluate(feed_dict)
        return bqm.energy(dict_binary_solution)

    @vartype_argument('vartype')
    def decode_solution(self, solution, vartype, feed_dict=None):
        """Returns decoded solution.
        
        Args:
            solution (list[bit]/dict[label, bit]/dict[index, bit]):
                The solution returned from solvers.
            
            vartype (:class:`dimod.Vartype`/str/set, optional):
                Variable type of the solution.
                Accepted input values:
                
                * :class:`.Vartype.SPIN`, ``'SPIN'``, ``{-1, 1}``
                * :class:`.Vartype.BINARY`, ``'BINARY'``, ``{0, 1}``
            
            feed_dict (dict[str, float]):
                Specify the placeholder values.
                
        Returns:
            tuple(dict, dict, float): Tuple of the decoded solution,
            broken constraints and energy.
            Structure of decoded_solution is defined by :obj:`structure`.
        """

        def put_value_with_keys(dict_body, keys, value):
            for key in keys[:-1]:
                if key not in dict_body:
                    dict_body[key] = {}
                dict_body = dict_body[key]
            dict_body[keys[-1]] = value

        decoded_solution = {}

        dict_bin_solution = self._parse_solution(solution, vartype)

        for label, bit in dict_bin_solution.items():
            if label in self.structure:
                if vartype == dimod.SPIN:
                    out_value = 2 * bit - 1
                elif vartype == dimod.BINARY:
                    out_value = bit
                put_value_with_keys(decoded_solution, self.structure[label], out_value)

        # Check satisfaction of constraints
        broken_const = {}
        for label, const in self.constraints.items():
            energy = const.energy(dict_bin_solution, feed_dict)
            if energy > 0.0:
                result_value = {var: dict_bin_solution[var] for var in
                                reduce(or_, [k.keys for k in const.polynomial.keys()])}
                broken_const[label] = {"result": result_value, "penalty": energy}
            elif energy < 0.0:
                raise ValueError("The energy of the constraint \"{label}\" is {energy}."
                                 "But an energy of constraints should not be negative."
                                 .format(label=label, energy=energy))

        problem_energy = self.energy(dict_bin_solution, dimod.BINARY, feed_dict)

        return decoded_solution, broken_const, problem_energy

    def decode_dimod_response(self, response, topk=None):
        """Decode the solution of :class:`dimod.Response`.
        
        For more details about :class:`dimod.Response`,
        see `dimod.Response 
        <https://dimod.readthedocs.io/en/latest/reference/response.html>`_.
        
        Args:
            response (:class:`dimod.Response`):
                The solution returned from dimod sampler.
            topk (int, default=None):
                Decode only top-k (energy is smaller) solutions.
        
        Returns:
            list[tuple(dict, dict, float)]: List of tuple of the decoded solution and
            broken constraints and energy. Solutions are sorted by energy.
            Structure of decoded_solution is defined by :obj:`structure`.
        """
        top_indices = np.argsort(response.record.energy)
        if topk:
            top_indices = top_indices[:topk]

        dict_solutions = list(dict(zip(response.variable_labels, sample))
                              for sample in response.record.sample[top_indices])
        decoded_solutions = []
        for sol in dict_solutions:
            decoded_solutions.append(self.decode_solution(sol, response.vartype))
        return decoded_solutions

    def to_dimod_bqm(self, feed_dict=None):
        """Returns :class:`dimod.BinaryQuadraticModel`.
        
        For more details about :class:`dimod.BinaryQuadraticModel`,
        see `dimod.BinaryQuadraticModel 
        <https://dimod.readthedocs.io/en/latest/reference/binary_quadratic_model.html>`_.
        
        Args:
            feed_dict (dict[str, float]):
                If the expression contains :class:`Placeholder` objects,
                you have to specify the value of them by :obj:`Placeholder`.
        
        Returns:
            :class:`dimod.BinaryQuadraticModel` with vartype set to `dimod.BINARY`.
        """
        return self.compiled_qubo.evaluate(feed_dict)

    def to_qubo(self, index_label=False, feed_dict=None):
        """Returns QUBO and energy offset.
        
        Args:
            index_label (bool):
                If true, the keys of returned QUBO are indexed with a positive integer number.
            
            feed_dict (dict[str, float]):
                If the expression contains :class:`Placeholder` objects,
                you have to specify the value of them by :obj:`Placeholder`.
        
        Returns:
            tuple(QUBO, float): Tuple of QUBO and energy offset.
            QUBO takes the form of ``dict[(label, label), value]``.
        
        Examples:
            This example creates the :obj:`model` from the expression, and
            we get the resulting QUBO by calling :func:`model.to_qubo()`.
            
            >>> from pyqubo import Qbit
            >>> x, y, z = Qbit("x"), Qbit("y"), Qbit("z")
            >>> model = (x*y + y*z + 3*z).compile()
            >>> pprint(model.to_qubo())
            ({('x', 'x'): 0.0,
              ('x', 'y'): 1.0,
              ('y', 'y'): 0.0,
              ('y', 'z'): 1.0,
              ('z', 'z'): 3.0},
             0.0)
            
            If you want a QUBO which has index labels, specify the argument ``index_label=True``.
            The mapping of the indices and the corresponding labels is
            stored in :obj:`model.variable_order`.
            
            >>> pprint(model.to_qubo(index_label=True))
            ({(0, 0): 0.0, (0, 1): 1.0, (1, 1): 0.0, (1, 2): 1.0, (2, 2): 3.0}, 0.0)
            >>> model.variable_order
            ['x', 'y', 'z']
            
        """

        bqm = self.compiled_qubo.evaluate(feed_dict)
        q, offset = bqm.to_qubo()

        # Convert label to index if index_label is true
        qubo = {}
        for (label1, label2), v in q.items():
            if index_label:
                i = self.label2index[label1]
                j = self.label2index[label2]
            else:
                i = label1
                j = label2
            qubo[(i, j)] = v

        return qubo, offset

    def to_ising(self, index_label=False, feed_dict=None):
        """Returns Ising Model and energy offset.

        Args:
            index_label (bool):
                If true, the keys of returned Ising model are
                indexed with a positive integer number.
            
            feed_dict (dict[str, float]):
                If the expression contains :class:`Placeholder` objects,
                you have to specify the value of them by :obj:`Placeholder`.

        Returns:
            tuple(linear, quadratic, float):
                Tuple of Ising Model and energy offset. Where `linear` takes the form of
                ``(dict[label, value])``, and `quadratic` takes the form of
                ``dict[(label, label), value]``.

        Examples:
            This example creates the :obj:`model` from the expression, and
            we get the resulting Ising model by calling :func:`model.to_ising()`.
            
            >>> from pyqubo import Qbit
            >>> x, y, z = Qbit("x"), Qbit("y"), Qbit("z")
            >>> model = (x*y + y*z + 3*z).compile()
            >>> pprint(model.to_ising())
            ({'x': 0.25, 'y': 0.5, 'z': 1.75}, {('x', 'y'): 0.25, ('y', 'z'): 0.25}, 2.0)
            
            If you want a Ising model which has index labels,
            specify the argument ``index_label=True``.
            The mapping of the indices and the corresponding labels is
            stored in :obj:`model.variable_order`.
            
            >>> pprint(model.to_ising(index_label=True))
            ({0: 0.25, 1: 0.5, 2: 1.75}, {(0, 1): 0.25, (1, 2): 0.25}, 2.0)
            >>> model.variable_order
            ['x', 'y', 'z']

        """

        bqm = self.compiled_qubo.evaluate(feed_dict)
        linear, quadratic, offset = bqm.to_ising()

        # Convert label to index if index_label is true
        new_linear = {}
        for label, v in linear.items():
            if index_label:
                i = self.label2index[label]
            else:
                i = label
            new_linear[i] = v

        new_quadratic = {}
        for (label1, label2), v in quadratic.items():
            if index_label:
                i = self.label2index[label1]
                j = self.label2index[label2]
            else:
                i = label1
                j = label2
            new_quadratic[(i, j)] = v

        return new_linear, new_quadratic, offset



# Copyright 2018 Recruit Communications Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc
import copy
from collections import defaultdict
from operator import or_, xor
from six.moves import reduce
import dimod
import six

from .coefficient import Coefficient
from .model import Model
from .compiled_qubo import CompiledQubo
from .compiled_constraint import CompiledConstraint
from .binaryprod import BinaryProd
from .placeholderprod import PlaceholderProd


@six.add_metaclass(abc.ABCMeta)
class Express:
    """Abstract class of pyqubo expression.
    
    All basic component class such as :class:`.Qbit`, :class:`.Spin` or :class:`.Add`
    inherits :class:`.Express`.
    
    .. graphviz::
    
        digraph {
            graph [size="2.5, 2.5"]
            node [shape=rl]
            add [label=AddList]
            qbit_a [label="Qbit(a)"]
            qbit_b [label="Qbit(b)"]
            mul_1 [label="Mul"]
            mul_2 [label="Mul"]
            num_1 [label="Num(1)"]
            num_2 [label="Num(2)"]
            add -> num_1
            add -> mul_1
            mul_1 -> mul_2
            mul_1 -> qbit_a
            mul_2 -> qbit_b
            mul_2 -> num_2
        }
    
    For example, an expression :math:`2ab+1` (where :math:`a, b` is :class:`Qbit` variable) is
    represented by the binary tree above.
    
    Note:
        This class is an abstract class of all component of expressions.
    
    Example:
        We write mathematical expressions with objects such as :class:`Qbit` or :class:`Spin`
        which inherit :class:`.Express`.
        
        >>> from pyqubo import Qbit
        >>> a, b = Qbit("a"), Qbit("b")
        >>> 2*a*b + 1
        (((Qbit(a)*Num(2))*Qbit(b))+Num(1))

    """

    CONST_TERM_KEY = BinaryProd(set())
    PROD_SYM = '*'

    def __init__(self):
        pass

    def __rmul__(self, other):
        """It is called when `other(number) - self`"""
        return self.__mul__(other)

    def __mul__(self, other):
        """It is called when `self * other(any object)`"""
        return Mul(self, other)

    def __rsub__(self, other):
        """It is called when `other(number) - self`"""
        return AddList([Mul(self, -1), other])

    def __sub__(self, other):
        """It is called when `self - other(any object)`"""
        if other == 0.0:
            return self
        else:
            if isinstance(other, Express):
                return self.__add__(Mul(other, -1))
            else:
                return self.__add__(-other)

    def __radd__(self, other):
        """It is called when `other(number) + self`"""
        return self.__add__(other)
    
    def __add__(self, other):
        """It is called when `self + other(any object)`"""
        if other == 0.0:
            return self
        else:
            return AddList([self, other])

    def __pow__(self, order):
        if int(order) == order and order >= 1:
            return reduce(lambda a, b: Mul(a, b), order * [self])
        else:
            raise ValueError("Power of {} th order cannot be done.".format(order))

    def __div__(self, other):
        """It is called when `self / other(any object)`"""
        if not isinstance(other, Express):
            return Mul(self, other ** -1)
        else:
            raise ValueError("Expression cannot be divided by Expression.")

    def __rdiv__(self, other):
        """It is called when `other(number) / self`"""
        raise ValueError("Expression cannot be divided by Expression.")

    def __truediv__(self, other):  # pragma: no cover
        """division in Python3"""
        return self.__div__(other)

    def __rtruediv__(self, other):  # pragma: no cover
        """It is called when `other(number) / self`"""
        return self.__rdiv__(other)

    def __neg__(self):
        """Negative value of expression."""
        return Mul(self, Num(-1))

    def __ne__(self, other):
        return not self.__eq__(other)

    @abc.abstractmethod
    def __repr__(self):  # pragma: no cover
        pass

    @abc.abstractmethod
    def __eq__(self, other):  # pragma: no cover
        pass

    @abc.abstractmethod
    def __hash__(self):  # pragma: no cover
        pass

    def compile(self, strength=5.0):
        """Returns the compiled :class:`Model`.
        
        This method reduces the degree of the expression if the degree is higher than 2,
        and convert it into :class:`.Model` which has information about QUBO.
        
        Args:
            strength (float): The strength of the reduction constraint.
                Insufficient strength can result in the binary quadratic model
                not having the same minimizations as the polynomial.

        Returns:
            :class:`Model`: The model compiled from the :class:`.Express`.
        
        Examples:
            In this example, there is a higher order term :math:`abcd`. It is decomposed as
            [[``a*d``, ``c``], ``b``] hierarchically and converted into QUBO.
            By calling :func:`to_qubo()` of the :obj:`model`, we get the resulting QUBO.
            
            >>> from pyqubo import Qbit
            >>> a, b, c, d = Qbit("a"), Qbit("b"), Qbit("c"), Qbit("d")
            >>> model = (a*b*c + a*b*d).compile()
            >>> pprint(model.to_qubo())
            ({('a', 'a'): 0.0,
              ('a', 'a*b'): -10.0,
              ('a', 'b'): 5.0,
              ('a*b', 'a*b'): 15.0,
              ('a*b', 'b'): -10.0,
              ('a*b', 'c'): 1.0,
              ('a*b', 'd'): 1.0,
              ('b', 'b'): 0.0,
              ('c', 'c'): 0.0,
              ('d', 'd'): 0.0},
             0.0)
        """
        def compile_param_if_express(val):
            if isinstance(val, Express):
                return val._compile_param()
            else:
                return val

        # Constraint for AND(multiplier, multiplicand) = product
        def binary_product(multiplier, multiplicand, product, weight):
            return {
                BinaryProd({product}): 3.0 * weight,
                BinaryProd({multiplicand, product}): -2.0 * weight,
                BinaryProd({multiplier, product}): -2.0 * weight,
                BinaryProd({multiplier, multiplicand}): weight
            }

        # When the label contains PROD_SYM, elements of products should be sorted
        # such that the resulting label is uniquely determined.
        def normalize_label(label):
            s = Express.PROD_SYM
            if s in label:
                return s.join(sorted(label.split(s)))
            else:
                return label

        # Expand the expression to polynomial
        expanded, constraints = Express._expand(self)

        # Make polynomials quadratic
        offset = 0.0
        pubo = {}
        for term_key, value in expanded.items():
            if term_key.is_constant():
                offset = value
            else:
                pubo[tuple(term_key.keys)] = value
        bqm = dimod.make_quadratic(pubo, strength, dimod.BINARY)
        bqm_qubo, bqm_offset = bqm.to_qubo()

        # Extracts product constrains
        product_consts = {}
        for (a, b), v in bqm.info['reduction'].items():
            prod = normalize_label(v['product'])
            a = normalize_label(a)
            b = normalize_label(b)
            product_consts["AND({},{})={}".format(a, b, prod)]\
                = binary_product(a, b, prod, strength)

        # Normalize labels and compile values of the QUBO
        compiled_qubo = {}
        for (label1, label2), value in bqm_qubo.items():
            norm_label1 = normalize_label(label1)
            norm_label2 = normalize_label(label2)

            # Sort the tuple of labels such that the created key is uniquely determined.
            if norm_label2 > norm_label1:
                label_key = (norm_label1, norm_label2)
            else:
                label_key = (norm_label2, norm_label1)
            # Compile values of the QUBO
            compiled_qubo[label_key] = compile_param_if_express(value)
        compiled_qubo = CompiledQubo(compiled_qubo, compile_param_if_express(offset + bqm_offset))

        # compile values of constraint
        compiled_constraints = {}
        for label, constraint in Express._merge_dict(constraints, product_consts).items():
            compiled_constraints[label] =\
                CompiledConstraint({prod: compile_param_if_express(value)
                                    for prod, value in constraint.items()})

        # Merge structures
        uniq_variables = Express._unique_vars(self)
        structure = reduce(Express._merge_dict, [var.structure for var in uniq_variables])

        return Model(compiled_qubo, structure, compiled_constraints)

    def _compile_param(self):
        expanded, _ = Express._expand_placeholder(self)
        return Coefficient(expanded)

    @staticmethod
    def _merge_dict(dict1, dict2):
        dict1_copy = copy.copy(dict1)
        dict1_copy.update(dict2)
        return dict1_copy

    @staticmethod
    def _merge_dict_update(const1, const2):
        const1.update(const2)
        return const1

    @staticmethod
    def _unique_vars(exp):
        if isinstance(exp, AddList):
            return reduce(or_, [Express._unique_vars(term) for term in exp.terms])
        elif isinstance(exp, Mul):
            return Express._unique_vars(exp.left) | Express._unique_vars(exp.right)
        elif isinstance(exp, Add):
            return Express._unique_vars(exp.left) | Express._unique_vars(exp.right)
        elif isinstance(exp, Placeholder):
            return set()
        elif isinstance(exp, Num):
            return set()
        elif isinstance(exp, Constraint):
            return Express._unique_vars(exp.child)
        elif isinstance(exp, Qbit):
            return {exp}
        elif isinstance(exp, Spin):
            return {exp}
        elif isinstance(exp, UserDefinedExpress):
            return Express._unique_vars(exp.express)
        else:
            raise TypeError("Unexpected input type {}.".format(type(exp)))  # pragma: no cover

    @staticmethod
    def _merge_term(term1, term2):
        if len(term1) < len(term2):
            for k, v in term1.items():
                term2[k] += v
            r = term2
        else:
            for k, v in term2.items():
                term1[k] += v
            r = term1
        return r

    @staticmethod
    def _expand_placeholder(exp):
        """Expand the placeholder expression hierarchically into dict format."""

        if isinstance(exp, AddList):
            expanded, const = reduce(
                lambda arg1, arg2:
                (Express._merge_term(arg1[0], arg2[0]), Express._merge_dict_update(arg1[1], arg2[1])),
                [Express._expand_placeholder(term) for term in exp.terms])
            return expanded, const

        elif isinstance(exp, Mul):
            left, left_const = Express._expand_placeholder(exp.left)
            right, right_const = Express._expand_placeholder(exp.right)
            expanded_terms = defaultdict(float)
            for k1, v1 in left.items():
                for k2, v2 in right.items():
                    merged_key = PlaceholderProd.merge_term_key(k1, k2)
                    expanded_terms[merged_key] += v1 * v2
            return expanded_terms, Express._merge_dict_update(left_const, right_const)

        elif isinstance(exp, Placeholder):
            expanded_terms = defaultdict(float)
            expanded_terms[PlaceholderProd({exp.label: 1.0})] = 1.0
            return expanded_terms, {}

        elif isinstance(exp, Num):
            terms = defaultdict(float)
            terms[PlaceholderProd({})] = exp.value
            return terms, {}

        else:
            raise TypeError("Unexpected input type {}.".format(type(exp)))  # pragma: no cover

    @staticmethod
    def _expand(exp):
        """Expand the expression hierarchically into dict format.
        
        For example, ``2*Qbit(a)*Qbit(b) + 1`` is represented as
        dict format ``{BinaryProd(ab): 2.0, CONST_TERM_KEY: 1.0}``.
        
        Let's see how this dict is created step by step.
        First, focus on `2*Qbit(a)*Qbit(b)` in which each expression is expanded as
        _expand(Num(2)) # => {CONST_TERM_KEY: 2.0}
        _expand(Qbit("a")) # => {BinaryProd(a): 1.0}
        _expand(Qbit("b")) # => {BinaryProd(b): 1.0}
        respectively.
        
        :class:`Mul` combines the expression in the following way
        _expand(Mul(Num(2), Qbit("a"))) # => {BinaryProd(a): 2.0}
        _expand(Mul(Qbit("b"), Mul(Num(2), Qbit("a")))) # => {BinaryProd(ab): 2.0}
        
        Finally, :class:`Add` combines the ``{BinaryProd(ab): 2.0}`` and `{CONST_TERM_KEY: 1.0}`,
        and we get the final form {BinaryProd(ab): 2.0, CONST_TERM_KEY: 1.0}
        
        Args:
            exp (:class:`Express`): Input expression.
        
        Returns:
            tuple(expanded_terms, constraints):
                tuple of expanded_terms and constrains. Expanded expression takes the form
                ``dict[:class:`BinaryProd`, value]``.
                Constraints takes the form of ``dict[label, expanded_terms]``.
        
        """

        if isinstance(exp, Add):
            left, left_const = Express._expand(exp.left)
            right, right_const = Express._expand(exp.right)
            result = Express._merge_term(right, left)
            return result, Express._merge_dict_update(left_const, right_const)

        elif isinstance(exp, AddList):
            expanded, const = reduce(
                lambda arg1, arg2:
                (Express._merge_term(arg1[0], arg2[0]), Express._merge_dict_update(arg1[1], arg2[1])),
                [Express._expand(term) for term in exp.terms])
            return expanded, const

        elif isinstance(exp, Mul):
            left, left_const = Express._expand(exp.left)
            right, right_const = Express._expand(exp.right)
            expanded_terms = defaultdict(float)
            for k1, v1 in left.items():
                for k2, v2 in right.items():
                    merged_key = BinaryProd.merge_term_key(k1, k2)
                    expanded_terms[merged_key] += v1 * v2
            return expanded_terms, Express._merge_dict_update(left_const, right_const)

        elif isinstance(exp, Placeholder):
            expanded_terms = defaultdict(float)
            expanded_terms[Express.CONST_TERM_KEY] = exp
            return expanded_terms, {}

        elif isinstance(exp, Constraint):
            child, child_const = Express._expand(exp.child)
            child_const[exp.label] = copy.copy(child)
            return child, child_const

        elif isinstance(exp, Num):
            terms = defaultdict(float)
            terms[Express.CONST_TERM_KEY] = exp.value
            return terms, {}

        elif isinstance(exp, Qbit):
            terms = defaultdict(float)
            terms[BinaryProd({exp.label})] = 1.0
            return terms, {}

        elif isinstance(exp, Spin):
            terms = defaultdict(float)
            terms[BinaryProd({exp.label})] = 2.0
            terms[Express.CONST_TERM_KEY] = -1.0
            return terms, {}

        elif isinstance(exp, UserDefinedExpress):
            return Express._expand(exp.express)

        else:
            raise TypeError("Unexpected input type {}.".format(type(exp)))  # pragma: no cover


class UserDefinedExpress(Express):
    """User Defined Express.
    
    User can define her/his own expression by inheriting :class:`UserDefinedExpress`.
    
    Attributes:
        express (Express): User can define an original expression by defining this member.
        
    Example:
        Define the :class:`LogicalAnd` class by inheriting :class:`UserDefinedExpress`.
        
        >>> from pyqubo import UserDefinedExpress
        >>> class LogicalAnd(UserDefinedExpress):
        ...     def __init__(self, bit_a, bit_b):
        ...         express = bit_a * bit_b
        ...         super(LogicalAnd, self).__init__(express)
    """

    def __init__(self, express):
        assert isinstance(express, Express)
        super(UserDefinedExpress, self).__init__()
        self.express = express

    def __hash__(self):
        return hash(self.express)

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        else:
            return self.express == other.express

    def __repr__(self):
        return "{}({})".format(self.__class__.__name__, self.express)


class Placeholder(Express):
    """Placeholder expression.
    
    You can specify the value of the :class:`Placeholder` when creating the QUBO.
    By using :class:`Placeholder`, you can change the value without compiling again.
    This is useful when you need to update the strength of constraint gradually.
    
    Args:
        label (str): The label of the placeholder.
    
    Example:
        The value of the placeholder is specified when you call :func:`to_qubo`.
        
        >>> from pyqubo import Qbit, Placeholder
        >>> x, y, a = Qbit('x'), Qbit('y'), Placeholder('a')
        >>> exp = a*x*y + 2*x
        >>> pprint(exp.compile().to_qubo(feed_dict={'a': 3}))
        ({('x', 'x'): 2.0, ('x', 'y'): 3.0, ('y', 'y'): 0.0}, 0.0)
        >>> pprint(exp.compile().to_qubo(feed_dict={'a': 5}))
        ({('x', 'x'): 2.0, ('x', 'y'): 5.0, ('y', 'y'): 0.0}, 0.0)
    """

    def __init__(self, label):
        super(Placeholder, self).__init__()
        self.label = label

    def __hash__(self):
        return hash(self.label)

    def __eq__(self, other):
        if not isinstance(other, Placeholder):
            return False
        else:
            return self.label == other.label

    def __repr__(self):
        return "Placeholder({})".format(self.label)


class Constraint(Express):
    """Constraint expression.
    
    You can specify the constraint part in your expression.
    
    Args:
        child (:class:`Express`): The expression you want to specify as a constraint.
        
        label (str): The label of the constraint. You can identify constraints by the label.
    
    Example:
        When the solution is broken, `decode_solution` can detect it.
        In this example, we introduce a constraint :math:`a+b=1`.
        
        >>> from pyqubo import Qbit, Constraint
        >>> a, b = Qbit('a'), Qbit('b')
        >>> exp = a + b + Constraint((a+b-1)**2, label="one_hot")
        >>> model = exp.compile()
        >>> sol, broken, energy = model.decode_solution({'a': 1, 'b': 1}, vartype='BINARY')
        >>> pprint(broken)
        {'one_hot': {'penalty': 1.0, 'result': {'a': 1, 'b': 1}}}
        >>> sol, broken, energy = model.decode_solution({'a': 1, 'b': 0}, vartype='BINARY')
        >>> pprint(broken)
        {}
    """

    def __init__(self, child, label):
        assert isinstance(label, str), "label should be string."
        assert isinstance(child, Express), "child should be an Express instance."
        super(Constraint, self).__init__()
        self.child = child
        self.label = label

    def __hash__(self):
        return hash(self.label) ^ hash(self.child)

    def __eq__(self, other):
        if not isinstance(other, Constraint):
            return False
        else:
            return self.label == other.label and self.child == other.child

    def __repr__(self):
        return "Const({}, {})".format(self.label, repr(self.child))


class Spin(Express):
    """Spin variable i.e. {-1, 1}.
    
    Args:
        label (str): The label of a variable. A variable is identified by this label.
        
        structure (dict/optional): Variable structure.
    
    Example:
        >>> from pyqubo import Spin
        >>> a, b = Spin('a'), Spin('b')
        >>> exp = 2*a*b + 3*a
        >>> pprint(exp.compile().to_qubo())
        ({('a', 'a'): 2.0, ('a', 'b'): 8.0, ('b', 'b'): -4.0}, -1.0)
    """

    def __init__(self, label, structure=None):
        assert isinstance(label, str), "label should be string."
        assert Express.PROD_SYM not in label, "label should not contain {}".format(Express.PROD_SYM)
        super(Spin, self).__init__()
        if not structure:
            # if no structure is given
            self.structure = {label: (label,)}
        else:
            self.structure = structure
        self.label = label

    def __repr__(self):
        return "Spin({})".format(self.label)

    def __hash__(self):
        return hash(self.label)

    def __eq__(self, other):
        """Returns whether the label is same or not."""
        if not isinstance(other, Spin):
            return False
        else:
            return self.label == other.label


class Qbit(Express):
    """Binary variable i.e. {0, 1}.
    
    Args:
        label (str): The label of a variable. A variable is identified by this label.
        
        structure (dict/optional): Variable structure.
    
    Example:
        >>> from pyqubo import Qbit
        >>> a, b = Qbit('a'), Qbit('b')
        >>> exp = 2*a*b + 3*a
        >>> pprint(exp.compile().to_qubo())
        ({('a', 'a'): 3.0, ('a', 'b'): 2.0, ('b', 'b'): 0.0}, 0.0)
    """

    def __init__(self, label, structure=None):
        assert isinstance(label, str), "Label should be string."
        assert Express.PROD_SYM not in label, "label should not contain {}".format(Express.PROD_SYM)
        super(Qbit, self).__init__()
        if not structure:
            # if no structure is given
            self.structure = {label: (label,)}
        else:
            self.structure = structure
        self.label = label

    def __repr__(self):
        return "Qbit({})".format(self.label)

    def __hash__(self):
        return hash(self.label)

    def __eq__(self, other):
        """Returns whether the label is same or not."""
        if not isinstance(other, Qbit):
            return False
        else:
            return self.label == other.label


class Mul(Express):
    """Product of expressions.
    
    Args:
        left (:class:`Express`): An expression
        
        right (:class:`Express`): An expression

    Example:        
        You can multiply expressions with either the built-in operator or :class:`Mul`.
        
        >>> from pyqubo import Qbit, Mul
        >>> a, b = Qbit('a'), Qbit('b')
        >>> a * b
        (Qbit(a)*Qbit(b))
        >>> Mul(a, b)
        (Qbit(a)*Qbit(b))
    """

    def __init__(self, left, right):
        super(Mul, self).__init__()
        # When right is a number (non-Express). (Only right can be a number.)
        if isinstance(left, Express) and not isinstance(right, Express):
            self.left = left
            self.right = Num(right)

        # When both arguments are Express.
        elif isinstance(left, Express) and isinstance(right, Express):
            self.left = left
            self.right = right
        else:
            raise ValueError("left should be an Express instance.")

    def __repr__(self):
        return "({}*{})".format(repr(self.left), repr(self.right))

    def __hash__(self):
        return hash(self.left) ^ hash(self.right)

    def __eq__(self, other):
        if not isinstance(other, Mul):
            return False
        elif self.left == other.left and self.right == other.right:
            return True
        elif self.right == other.left and self.left == other.right:
            return True
        else:
            return False


class Add(Express):
    """Addition of expressions (deprecated).
    
    Args:
        left (:class:`Express`): An expression
        
        right (:class:`Express`): An expression
    
    Example:
        You can add expressions with either the built-in operator or :class:`Add`.
        
        >>> from pyqubo import Qbit, Add
        >>> a, b = Qbit('a'), Qbit('b')
        >>> a + b
        (Qbit(a)+Qbit(b))
        >>> Add(a, b)
        (Qbit(a)+Qbit(b))
    
    """

    def __init__(self, left, right):
        super(Add, self).__init__()
        # When right is a number (non-Express). (Only right can be a number.)
        if isinstance(left, Express) and not isinstance(right, Express):
            self.left = left
            self.right = Num(right)

        # When both arguments are Express.
        elif isinstance(left, Express) and isinstance(right, Express):
            self.left = left
            self.right = right
        else:
            raise ValueError("left should be an Express instance.")

    def __hash__(self):
        return hash(self.left) ^ hash(self.right)

    def __repr__(self):
        return "({}+{})".format(repr(self.left), repr(self.right))

    def __eq__(self, other):
        if not isinstance(other, Add):
            return False
        elif self.left == other.left and self.right == other.right:
            return True
        elif self.right == other.left and self.left == other.right:
            return True
        else:
            return False


class AddList(Express):
    """Addition of a list of expressions.
    
    Args:
        terms (list[:class:`Express`]): a list of expressions
    
    Example:
        You can add expressions with either the built-in operator or :class:`AddList`.
        
        >>> from pyqubo import Qbit, AddList
        >>> a, b = Qbit('a'), Qbit('b')
        >>> a + b
        (Qbit(a)+Qbit(b))
        >>> AddList([a, b])
        (Qbit(a)+Qbit(b))
    """

    def __init__(self, terms):
        super(AddList, self).__init__()
        new_terms = []
        for term in terms:
            if isinstance(term, Express):
                new_terms.append(term)
            else:
                new_terms.append(Num(term))
        self.terms = new_terms

    def __add__(self, other):
        """ Override __add__().
        
        To prevent a deep nested expression, __add__() returns AddList object
        where the added expression object is appended to :obj:`terms`.
        
        Returns:
            :class:`AddList`
        """
        if other == 0.0:
            return self
        else:
            if not isinstance(other, Express):
                other = Num(other)
            return AddList(self.terms + [other])

    def __hash__(self):
        return reduce(xor, [hash(term) for term in self.terms])

    def __repr__(self):
        return "({})".format("+".join([repr(e) for e in self.terms]))

    def __eq__(self, other):
        if not isinstance(other, AddList):
            return False
        else:
            return set(self.terms) == set(other.terms)


class Num(Express):
    """Expression of number
    
    Args:
        value (float): the value of the number.
    
    Example:
        >>> from pyqubo import Qbit, Num
        >>> a = Qbit('a')
        >>> a + 1
        (Qbit(a)+Num(1))
        >>> a + Num(1)
        (Qbit(a)+Num(1))
    """
    def __init__(self, value):
        super(Num, self).__init__()
        self.value = value

    def __hash__(self):
        return hash(self.value)

    def __repr__(self):
        return "Num({})".format(str(self.value))

    def __eq__(self, other):
        if not isinstance(other, Num):
            return False
        else:
            return self.value == other.value




# Copyright 2018 Recruit Communications Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
from pyqubo import Coefficient, PlaceholderProd


class TestCoefficient(unittest.TestCase):

    def test_coefficient_exception(self):
        coeff = Coefficient({PlaceholderProd({'a': 1, 'b': 1}): 2.0, PlaceholderProd({}): 2.0})
        self.assertRaises(ValueError, lambda: coeff.evaluate({}))
        self.assertRaises(ValueError, lambda: coeff.evaluate({'a': 1.0}))



# Copyright 2018 Recruit Communications Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

from pyqubo import Qbit, Spin, AddList, Mul, Add, Num, Placeholder, Constraint


class TestExpressEquality(unittest.TestCase):

    def test_equality_of_add_list(self):
        exp1 = AddList([Qbit("a"), Qbit("b")])
        exp2 = AddList([Qbit("b"), Qbit("a")])
        exp3 = AddList([Qbit("a"), Qbit("a")])
        self.assertTrue(exp1 == exp2)
        self.assertTrue(hash(exp1) == hash(exp2))
        self.assertFalse(exp1 == exp3)
        self.assertFalse(exp1 == Qbit("a"))

    def test_equality_of_add(self):
        exp1 = Add(Qbit("a"), Qbit("b"))
        exp2 = Add(Qbit("b"), Qbit("a"))
        exp3 = Add(Qbit("a"), Qbit("a"))
        exp4 = Add(Qbit("a"), Qbit("b"))
        exp5 = Add(Qbit("a"), 1)
        self.assertTrue(exp1 == exp2)
        self.assertTrue(exp1 == exp4)
        self.assertTrue(hash(exp1) == hash(exp2))
        self.assertTrue(exp1 != exp3)
        self.assertFalse(exp1 == exp5)
        self.assertFalse(exp1 == Qbit("a"))
        self.assertFalse(exp1 == exp3)
        self.assertEqual(repr(exp1), "(Qbit(a)+Qbit(b))")

    def test_equality_of_mul(self):
        exp1 = Mul(Qbit("a"), Qbit("b"))
        exp2 = Mul(Qbit("b"), Qbit("a"))
        exp3 = Mul(Qbit("a"), Qbit("a"))
        self.assertTrue(exp1 == exp2)
        self.assertTrue(hash(exp1) == hash(exp2))
        self.assertFalse(exp1 == exp3)
        self.assertTrue(exp1 != Qbit("a"))

    def test_equality_of_num(self):
        self.assertTrue(Num(1) == Num(1))
        self.assertFalse(Num(1) == Num(2))
        self.assertFalse(Num(1) == Qbit("a"))

    def test_equality_of_placeholder(self):
        p1 = Placeholder("p1")
        p2 = Placeholder("p2")
        p3 = Placeholder("p1")
        self.assertTrue(p1 == p3)
        self.assertTrue(hash(p1) == hash(p3))
        self.assertTrue(p1 != p2)
        self.assertTrue(p1 != Qbit("a"))

    def test_equality_of_const(self):
        c1 = Constraint(Qbit("a"), label="c1")
        c2 = Constraint(Qbit("b"), label="c1")
        c3 = Constraint(Qbit("a"), label="c3")
        c4 = Constraint(Qbit("a"), label="c1")
        self.assertTrue(c1 == c4)
        self.assertFalse(c1 != c4)
        self.assertTrue(hash(c1) == hash(c4))
        self.assertTrue(c1 != c2)
        self.assertTrue(c1 != c3)
        self.assertTrue(c1 != Qbit("a"))

    def test_equality_of_spin(self):
        a, b, c = Spin("a"), Spin("b"), Spin("a")
        self.assertTrue(a == c)
        self.assertFalse(a != c)
        self.assertTrue(hash(a) == hash(c))
        self.assertTrue(a != b)
        self.assertTrue(a != Qbit("a"))
        self.assertEqual(repr(a), "Spin(a)")

    def test_equality_of_qbit(self):
        a, b, c = Qbit("a"), Qbit("b"), Qbit("a")
        self.assertTrue(a == c)
        self.assertFalse(a != c)
        self.assertTrue(hash(a) == hash(c))
        self.assertTrue(a != b)
        self.assertTrue(a != Spin("a"))

    def test_equality_of_express(self):
        a, b = Qbit("a"), Qbit("b")
        exp = a * b + 2*a - 1
        expected_exp = AddList([Mul(a, b), Num(-1.0), Mul(a, 2)])
        self.assertTrue(exp == expected_exp)

    def test_equality_sub(self):
        a, b = Qbit("a"), Qbit("b")
        exp = 1-a-b
        expected_exp = AddList([Mul(a, -1), Num(1.0), Mul(b, -1)])
        self.assertTrue(exp == expected_exp)
        self.assertTrue(exp - 0.0 == expected_exp)

    def test_equality_sub2(self):
        a, b = Qbit("a"), Qbit("b")
        exp = a-b-1
        expected_exp = AddList([a, Num(-1.0), Mul(b, -1)])
        self.assertTrue(exp == expected_exp)

    def test_equality_of_express_with_placeholder(self):
        a, b, p = Qbit("a"), Qbit("b"), Placeholder("p")
        exp = a + b - 1 + a * p
        expected_exp = AddList([a, Num(-1.0), b, Mul(p, a)])
        self.assertTrue(exp == expected_exp)

    def test_equality_of_express_with_const(self):
        a, b = Qbit("a"), Spin("b")
        exp = a + b - 1 + Constraint(a * b, label="const")
        expected_exp = AddList([a, Num(-1.0), b, Constraint(Mul(a, b), label="const")])
        self.assertTrue(exp == expected_exp)

    def test_repr(self):
        a, b, p = Qbit("a"), Qbit("b"), Placeholder("p")
        exp = a + p - 1 + Constraint(a * b, label="const")
        expected = "(Qbit(a)+Placeholder(p)+Num(-1)+Const(const, (Qbit(a)*Qbit(b))))"
        self.assertTrue(repr(exp) == expected)

    def test_express_error(self):
        self.assertRaises(ValueError, lambda: 2 / Qbit("a"))
        self.assertRaises(ValueError, lambda: Qbit("a") / Qbit("a"))
        self.assertRaises(ValueError, lambda: Qbit("a") ** 1.5)
        self.assertRaises(ValueError, lambda: Mul(1, Qbit("b")))
        self.assertRaises(ValueError, lambda: Add(1, Qbit("b")))



# Copyright 2018 Recruit Communications Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

from pyqubo import BinaryProd


class TestBinaryProd(unittest.TestCase):

    def test_equality(self):
        term_key1 = BinaryProd({"a", "b"})
        term_key2 = BinaryProd({"a", "b"})
        term_key3 = BinaryProd({"a", "b", "c"})
        self.assertTrue(term_key1 == term_key2)
        self.assertTrue(term_key1 != term_key3)
        self.assertTrue(term_key1 != 1)

    def test_equality_const(self):
        term_key1 = BinaryProd(set())
        term_key2 = BinaryProd(set())
        term_key3 = BinaryProd({"a"})
        self.assertTrue(term_key1.is_constant())
        self.assertEqual(term_key1, term_key2)
        self.assertNotEqual(term_key1, term_key3)

    def test_merge(self):
        term_key1 = BinaryProd({"a", "b"})
        term_key2 = BinaryProd({"a", "b"})
        term_key3 = BinaryProd({"b", "c"})
        term_key4 = BinaryProd({"a", "b", "c"})
        self.assertEqual(term_key1, BinaryProd.merge_term_key(term_key1, term_key2))
        self.assertEqual(term_key4, BinaryProd.merge_term_key(term_key2, term_key3))

    def test_evaluate(self):
        term_key1 = BinaryProd({"a", "b"})
        empty_key = BinaryProd(set())
        dict_value = {"a": 3.0, "b": 5.0}
        prod = term_key1.calc_product(dict_value)
        expected_prod = 15
        self.assertEqual(expected_prod, prod)
        self.assertEqual(1, empty_key.calc_product({}))

    def test_repr(self):
        term_key1 = BinaryProd({"a", "b"})
        empty_key = BinaryProd(set())
        self.assertIn(repr(term_key1), "a*b")
        self.assertEqual(repr(empty_key), "const")



# Copyright 2018 Recruit Communications Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

from pyqubo import solve_qubo, solve_ising, Spin


class TestSolver(unittest.TestCase):

    @staticmethod
    def create_number_partition_model():
        s1, s2, s3 = Spin("s1"), Spin("s2"), Spin("s3")
        H = (2 * s1 + 4 * s2 + 6 * s3) ** 2
        return H.compile()

    def test_solve_qubo(self):
        model = TestSolver.create_number_partition_model()
        qubo, offset = model.to_qubo()
        solution = solve_qubo(qubo, num_reads=1, sweeps=10)
        self.assertTrue(solution == {'s1': 0, 's2': 0, 's3': 1} or {'s1': 1, 's2': 1, 's3': 0})

    def test_solve_ising(self):
        model = TestSolver.create_number_partition_model()
        linear, quad, offset = model.to_ising()
        solution = solve_ising(linear, quad, num_reads=1, sweeps=10)
        self.assertTrue(solution == {'s1': -1, 's2': -1, 's3': 1} or {'s1': 1, 's2': 1, 's3': -1})



# Copyright 2018 Recruit Communications Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
import itertools
from pyqubo import Qbit, And, Or, Not, Xor


class TestLogic(unittest.TestCase):

    def test_and(self):
        a, b = Qbit('a'), Qbit('b')
        exp = And(a, b)
        model = exp.compile()
        for a, b in itertools.product(*[(0, 1)] * 2):
            e = int(model.energy({'a': a, 'b': b}, vartype='BINARY'))
            self.assertEqual(a*b, e)

    def test_or(self):
        a, b = Qbit('a'), Qbit('b')
        exp = Or(a, b)
        model = exp.compile()
        for a, b in itertools.product(*[(0, 1)] * 2):
            e = int(model.energy({'a': a, 'b': b}, vartype='BINARY'))
            self.assertEqual(int(a+b > 0), e)

    def test_not(self):
        a = Qbit('a')
        exp = Not(a)
        model = exp.compile()
        for a in [0, 1]:
            e = int(model.energy({'a': a}, vartype='BINARY'))
            self.assertEqual(1-a, e)

        self.assertEqual(repr(exp), "Not(((Qbit(a)*Num(-1))+Num(1)))")

    def test_xor(self):
        a, b = Qbit('a'), Qbit('b')
        exp = Xor(a, b)
        model = exp.compile()
        for a, b in itertools.product(*[(0, 1)] * 2):
            e = int(model.energy({'a': a, 'b': b}, vartype='BINARY'))
            if (a == 1 and b == 0) or (a == 0 and b == 1):
                self.assertTrue(e == 1)
            else:
                self.assertTrue(e == 0)



# Copyright 2018 Recruit Communications Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

from pyqubo import Qbit, AndConst, OrConst, XorConst, NotConst


class TestConstraint(unittest.TestCase):

    def test_not(self):
        a, b = Qbit("a"), Qbit("b")
        exp = NotConst(a, b, label="not")
        model = exp.compile()
        self.assertTrue(model.energy({"a": 1, "b": 0}, vartype="BINARY") == 0)
        self.assertTrue(model.energy({"a": 0, "b": 1}, vartype="BINARY") == 0)
        self.assertTrue(model.energy({"a": 1, "b": 1}, vartype="BINARY") > 0)
        self.assertTrue(model.energy({"a": 0, "b": 0}, vartype="BINARY") > 0)

    def test_and(self):
        a, b, c = Qbit("a"), Qbit("b"), Qbit("c")
        exp = AndConst(a, b, c, label="and")
        model = exp.compile()
        self.assertTrue(model.energy({"a": 1, "b": 1, "c": 1}, vartype="BINARY") == 0)
        self.assertTrue(model.energy({"a": 1, "b": 0, "c": 0}, vartype="BINARY") == 0)
        self.assertTrue(model.energy({"a": 0, "b": 1, "c": 0}, vartype="BINARY") == 0)
        self.assertTrue(model.energy({"a": 0, "b": 0, "c": 0}, vartype="BINARY") == 0)
        self.assertTrue(model.energy({"a": 1, "b": 1, "c": 0}, vartype="BINARY") > 0)
        self.assertTrue(model.energy({"a": 0, "b": 0, "c": 1}, vartype="BINARY") > 0)
        self.assertTrue(model.energy({"a": 1, "b": 0, "c": 1}, vartype="BINARY") > 0)
        self.assertTrue(model.energy({"a": 0, "b": 1, "c": 1}, vartype="BINARY") > 0)

    def test_or(self):
        a, b, c = Qbit("a"), Qbit("b"), Qbit("c")
        exp = OrConst(a, b, c, label="or")
        model = exp.compile()
        self.assertTrue(model.energy({"a": 1, "b": 1, "c": 1}, vartype="BINARY") == 0)
        self.assertTrue(model.energy({"a": 1, "b": 0, "c": 1}, vartype="BINARY") == 0)
        self.assertTrue(model.energy({"a": 0, "b": 1, "c": 1}, vartype="BINARY") == 0)
        self.assertTrue(model.energy({"a": 0, "b": 0, "c": 0}, vartype="BINARY") == 0)
        self.assertTrue(model.energy({"a": 1, "b": 1, "c": 0}, vartype="BINARY") > 0)
        self.assertTrue(model.energy({"a": 1, "b": 0, "c": 0}, vartype="BINARY") > 0)
        self.assertTrue(model.energy({"a": 0, "b": 1, "c": 0}, vartype="BINARY") > 0)
        self.assertTrue(model.energy({"a": 0, "b": 0, "c": 1}, vartype="BINARY") > 0)

    def test_xor(self):
        a, b, c = Qbit("a"), Qbit("b"), Qbit("c")
        exp = XorConst(a, b, c, label="xor")
        model = exp.compile()
        self.assertTrue(model.energy({"a": 1, "b": 1, "c": 0, "aux_xor": 1}, vartype="BINARY") == 0)
        self.assertTrue(model.energy({"a": 1, "b": 0, "c": 1, "aux_xor": 0}, vartype="BINARY") == 0)
        self.assertTrue(model.energy({"a": 0, "b": 1, "c": 1, "aux_xor": 0}, vartype="BINARY") == 0)
        self.assertTrue(model.energy({"a": 1, "b": 1, "c": 0, "aux_xor": 1}, vartype="BINARY") == 0)
        self.assertTrue(model.energy({"a": 0, "b": 0, "c": 1, "aux_xor": 1}, vartype="BINARY") > 0)
        self.assertTrue(model.energy({"a": 1, "b": 1, "c": 1, "aux_xor": 1}, vartype="BINARY") > 0)

    def test_equality(self):
        xor1 = XorConst(Qbit("a"), Qbit("b"), Qbit("c"), label="xor")
        xor2 = XorConst(Qbit("a"), Qbit("b"), Qbit("c"), label="xor")
        xor3 = XorConst(Qbit("b"), Qbit("c"), Qbit("a"), label="xor")
        or1 = OrConst(Qbit("a"), Qbit("b"), Qbit("c"), label="xor")
        self.assertTrue(xor1 + 1 == xor2 + 1)
        self.assertTrue(xor1 == xor2)
        self.assertFalse(xor1 == or1)
        self.assertFalse(xor1 == xor3)



# Copyright 2018 Recruit Communications Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

from pyqubo import Qbit, Constraint, Placeholder, Array
from pyqubo import assert_qubo_equal


class TestModel(unittest.TestCase):

    def test_to_qubo(self):
        a, b = Qbit("a"), Qbit("b")
        exp = 1 + a * b + a - 2
        model = exp.compile()
        qubo, offset = model.to_qubo()
        assert_qubo_equal(qubo, {("a", "a"): 1.0, ("a", "b"): 1.0, ("b", "b"): 0.0})
        self.assertTrue(offset == -1)

    def test_to_qubo_with_index(self):
        a, b = Qbit("a"), Qbit("b")
        exp = 1 + a * b + a - 2
        model = exp.compile()
        qubo, offset = model.to_qubo(index_label=True)
        assert_qubo_equal(qubo, {(0, 0): 1.0, (0, 1): 1.0, (1, 1): 0.0})
        self.assertTrue(offset == -1)

    def test_to_ising(self):
        a, b = Qbit("a"), Qbit("b")
        exp = 1 + a * b + a - 2
        model = exp.compile()
        linear, quad, offset = model.to_ising()
        self.assertTrue(linear == {'a': 0.75, 'b': 0.25})
        assert_qubo_equal(quad, {('a', 'b'): 0.25})
        self.assertTrue(offset == -0.25)

    def test_to_ising_with_index(self):
        a, b = Qbit("a"), Qbit("b")
        exp = 1 + a * b + a - 2
        model = exp.compile()
        linear, quad, offset = model.to_ising(index_label=True)
        self.assertTrue(linear == {0: 0.75, 1: 0.25})
        assert_qubo_equal(quad, {(0, 1): 0.25})
        self.assertTrue(offset == -0.25)

    def test_decode(self):
        x = Array.create("x", (2, 2), vartype="BINARY")
        exp = Constraint((x[1, 1] - x[0, 1]) ** 2, label="const")
        model = exp.compile()

        # check __repr__ of CompiledQubo
        expected_repr = "CompiledQubo({('x[0][1]', 'x[0][1]'): 1.0,\n" \
            " ('x[0][1]', 'x[1][1]'): -2.0,\n" \
            " ('x[1][1]', 'x[1][1]'): 1.0}, offset=0.0)"
        self.assertEqual(repr(model.compiled_qubo), expected_repr)

        # check __repr__ of Model
        expected_repr = "Model(CompiledQubo({('x[0][1]', 'x[0][1]'): 1.0,\n"\
                        " ('x[0][1]', 'x[1][1]'): -2.0,\n"\
                        " ('x[1][1]', 'x[1][1]'): 1.0}, offset=0.0), "\
                        "structure={'x[0][1]': ('x', 0, 1), 'x[1][1]': ('x', 1, 1)})"
        self.assertEqual(repr(model), expected_repr)

        # when the constraint is not broken

        # type of solution is dict[label, bit]
        decoded_sol, broken, energy = model.decode_solution({'x[0][1]': 1.0, 'x[1][1]': 1.0},
                                                            vartype="BINARY")
        self.assertTrue(decoded_sol == {'x': {0: {1: 1}, 1: {1: 1}}})
        self.assertTrue(len(broken) == 0)
        self.assertTrue(energy == 0)

        # type of solution is list[bit]
        decoded_sol, broken, energy = model.decode_solution([1, 1], vartype="BINARY")
        self.assertTrue(decoded_sol == {'x': {0: {1: 1}, 1: {1: 1}}})
        self.assertTrue(len(broken) == 0)
        self.assertTrue(energy == 0)

        # type of solution is dict[index_label(int), bit]
        decoded_sol, broken, energy = model.decode_solution({0: 1.0, 1: 1.0},
                                                            vartype="BINARY")
        self.assertTrue(decoded_sol == {'x': {0: {1: 1}, 1: {1: 1}}})
        self.assertTrue(len(broken) == 0)
        self.assertTrue(energy == 0)

        # when the constraint is broken

        # type of solution is dict[label, bit]
        decoded_sol, broken, energy = model.decode_solution({'x[0][1]': 0.0, 'x[1][1]': 1.0},
                                                            vartype="BINARY")
        self.assertTrue(decoded_sol == {'x': {0: {1: 0}, 1: {1: 1}}})
        self.assertTrue(len(broken) == 1)
        self.assertTrue(energy == 1)

        # invalid solution
        self.assertRaises(ValueError, lambda: model.decode_solution([1, 1, 1], vartype="BINARY"))
        self.assertRaises(ValueError, lambda: model.decode_solution(
            {'x[0][2]': 1.0, 'x[1][1]': 0.0}, vartype="BINARY"))
        self.assertRaises(TypeError, lambda: model.decode_solution((1, 1), vartype="BINARY"))

        # type of solution is dict[label, spin]
        decoded_sol, broken, energy = model.decode_solution({'x[0][1]': 1.0, 'x[1][1]': -1.0},
                                                            vartype="SPIN")
        self.assertTrue(decoded_sol == {'x': {0: {1: 1}, 1: {1: -1}}})
        self.assertTrue(len(broken) == 1)
        self.assertTrue(energy == 1)

    def test_decode2(self):
        x = Array.create("x", (2, 2), vartype="BINARY")
        exp = Constraint(-(x[1, 1] - x[0, 1]) ** 2, label="const")
        model = exp.compile()
        self.assertRaises(ValueError, lambda: model.decode_solution([1, 0], vartype="BINARY"))

    def test_work_with_dimod(self):
        import numpy as np
        import dimod
        n = 10
        A = [np.random.randint(-50, 50) for _ in range(n)]
        S = Array.create('S', n, "BINARY")
        H = sum(A[i] * S[i] for i in range(n)) ** 2
        model = H.compile()
        binary_bqm = model.to_dimod_bqm()
        sa = dimod.reference.SimulatedAnnealingSampler()
        response = sa.sample(binary_bqm, num_reads=10, num_sweeps=5)
        decoded_solutions = model.decode_dimod_response(response, topk=2)
        for (decoded_solution, broken, energy) in decoded_solutions:
            assert isinstance(decoded_solution, dict)
            assert isinstance(broken, dict)
            assert isinstance(energy, float)

        spin_bqm = binary_bqm.change_vartype(dimod.SPIN)
        response = sa.sample(spin_bqm, num_reads=10, num_sweeps=5)
        decoded_solutions = model.decode_dimod_response(response, topk=2)
        for (decoded_solution, broken, energy) in decoded_solutions:
            assert isinstance(decoded_solution, dict)
            assert isinstance(broken, dict)
            assert isinstance(energy, float)

    def test_placeholders(self):
        a, b, p = Qbit("a"), Qbit("b"), Placeholder("p")
        feed_dict = {'p': 2.0}
        exp = p * (1 + a * b + a)
        model = exp.compile()
        qubo, offset = model.to_qubo(index_label=False, feed_dict=feed_dict)
        dict_solution = {'a': 1, 'b': 0}
        dict_energy = model.energy(dict_solution, vartype="BINARY", feed_dict=feed_dict)
        list_solution = [1, 0]
        list_energy = model.energy(list_solution, vartype="BINARY", feed_dict=feed_dict)
        assert_qubo_equal(qubo, {('a', 'b'): 2.0, ('a', 'a'): 2.0, ('b', 'b'): 0.0})
        self.assertEqual(offset, 2.0)
        self.assertTrue(dict_energy, 2.0)
        self.assertTrue(list_energy, 2.0)

    def test_placeholder_in_constraint(self):
        a = Qbit("a")
        exp = Constraint(2 * Placeholder("c") + a, "const1")
        m = exp.compile()
        sol, broken, energy = m.decode_solution({"a": 1}, vartype="BINARY", feed_dict={"c": 1})
        self.assertEqual(energy, 3.0)
        self.assertEqual(broken, {'const1': {'result': {'a': 1}, 'penalty': 3.0}})
        self.assertEqual(sol, {'a': 1})



# Copyright 2018 Recruit Communications Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


def assert_qubo_equal(qubo1, qubo2):
    """ Assert the given QUBOs are identical.
    
    Args:
        qubo1 (dict[(label, label), float]): QUBO to be compared.
        
        qubo2 (dict[(label, label), float]): QUBO to be compared.
    """

    msg = "QUBO should be an dict instance."
    assert isinstance(qubo1, dict), msg
    assert isinstance(qubo2, dict), msg

    assert len(qubo1) == len(qubo2), "Number of elements in QUBO doesn't match."

    for (label1, label2), value in qubo1.items():
        if (label1, label2) in qubo2:
            if qubo2[label1, label2] != value:
                assert qubo2[label1, label2] == value,\
                    "Value of {key} doesn't match".format(key=(label1, label2))
        elif (label2, label1) in qubo2:
            if qubo2[label2, label1] != value:
                assert qubo2[label2, label1] == value, \
                    "Value of {key} doesn't match".format(key=(label2, label1))
        else:
            raise AssertionError("Key: {key} of qubo1 isn't contained in qubo2"
                                 .format(key=(label1, label2)))



# Copyright 2018 Recruit Communications Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

from pyqubo import Qbit, Spin, Placeholder, Constraint, Array, Add, Sum
from pyqubo import assert_qubo_equal


class TestExpressCompile(unittest.TestCase):

    def compile_check(self, exp, expected_qubo, expected_offset, expected_structure,
                      feed_dict=None):
        model = exp.compile(strength=5)
        qubo, offset = model.to_qubo(feed_dict=feed_dict)
        assert_qubo_equal(qubo, expected_qubo)
        self.assertEqual(qubo, expected_qubo)
        self.assertEqual(offset, expected_offset)
        self.assertEqual(model.structure, expected_structure)

    def test_compile(self):
        a, b = Qbit("a"), Qbit("b")
        exp = 1 + a*b + a - 2
        expected_qubo = {('a', 'a'): 1.0, ('a', 'b'): 1.0, ('b', 'b'): 0.0}
        expected_offset = -1
        expected_structure = {'a': ('a',), 'b': ('b',)}
        self.compile_check(exp, expected_qubo, expected_offset, expected_structure)

    def test_compile_expand(self):
        a, b = Qbit("a"), Qbit("b")
        exp = (a+b)*(a-b)
        expected_qubo = {('a', 'a'): 1.0, ('a', 'b'): 0.0, ('b', 'b'): -1.0}
        expected_offset = 0.0
        expected_structure = {'a': ('a',), 'b': ('b',)}
        self.compile_check(exp, expected_qubo, expected_offset, expected_structure)

    def test_compile_spin(self):
        a, b = Qbit("a"), Spin("b")
        exp = a * b
        expected_qubo = {('a', 'a'): -1.0, ('a', 'b'): 2.0, ('b', 'b'): 0.0}
        expected_offset = 0.0
        expected_structure = {'a': ('a',), 'b': ('b',)}
        self.compile_check(exp, expected_qubo, expected_offset, expected_structure)

    def test_compile_2nd_order(self):
        a, b = Qbit("a"), Qbit("b")
        exp = (Add(a, b)-3)**2
        expected_qubo = {('a', 'a'): -5.0, ('a', 'b'): 2.0, ('b', 'b'): -5.0}
        expected_offset = 9.0
        expected_structure = {'a': ('a',), 'b': ('b',)}
        self.compile_check(exp, expected_qubo, expected_offset, expected_structure)

    def test_compile_3rd_order(self):
        a, b = Qbit("a"), Qbit("b")
        exp = (a+b-2)**3
        expected_qubo = {('a', 'a'): 7.0, ('a', 'b'): -6.0, ('b', 'b'): 7.0}
        expected_offset = -8.0
        expected_structure = {'a': ('a',), 'b': ('b',)}
        self.compile_check(exp, expected_qubo, expected_offset, expected_structure)

    def test_compile_reduce_degree(self):
        a, b, c, d = Qbit("a"), Qbit("b"), Qbit("c"), Qbit("d")
        exp = a * b * c + b * c * d
        expected_qubo = {
            ('a', 'a'): 0.0,
            ('a', 'b*c'): 1.0,
            ('b', 'b'): 0.0,
            ('b', 'b*c'): -10.0,
            ('b', 'c'): 5.0,
            ('c', 'c'): 0.0,
            ('b*c', 'c'): -10.0,
            ('b*c', 'b*c'): 15.0,
            ('b*c', 'd'): 1.0,
            ('d', 'd'): 0.0}
        expected_offset = 0.0
        expected_structure = {'a': ('a',), 'b': ('b',), 'c': ('c',), 'd': ('d',)}

        self.compile_check(exp, expected_qubo, expected_offset, expected_structure)

    def test_compile_div(self):
        a, b = Qbit("a"), Qbit("b")
        exp = (a+b-2)/2
        expected_qubo = {('a', 'a'): 0.5, ('b', 'b'): 0.5}
        expected_offset = -1
        expected_structure = {'a': ('a',), 'b': ('b',)}
        self.compile_check(exp, expected_qubo, expected_offset, expected_structure)

    def test_compile_placeholder(self):
        a, b, w, v = Qbit("a"), Qbit("b"), Placeholder("w"), Placeholder("v")
        exp = w*(a+b-2) + v
        expected_qubo = {('a', 'a'): 3.0, ('b', 'b'): 3.0}
        expected_offset = -1
        expected_structure = {'a': ('a',), 'b': ('b',)}
        self.compile_check(exp, expected_qubo, expected_offset, expected_structure,
                           feed_dict={"w": 3.0, "v": 5.0})

    def test_compile_placeholder2(self):
        a, b, w, v = Qbit("a"), Qbit("b"), Placeholder("w"), Placeholder("v")
        exp = v*w*(a+b-2) + v
        expected_qubo = {('a', 'a'): 15.0, ('b', 'b'): 15.0}
        expected_offset = -25
        expected_structure = {'a': ('a',), 'b': ('b',)}
        self.compile_check(exp, expected_qubo, expected_offset, expected_structure,
                           feed_dict={"w": 3.0, "v": 5.0})

    def test_compile_placeholder3(self):
        a, v = Qbit("a"), Placeholder("v")
        exp = v*v*a + v
        expected_qubo = {('a', 'a'): 25.0}
        expected_offset = 5
        expected_structure = {'a': ('a',)}
        self.compile_check(exp, expected_qubo, expected_offset, expected_structure,
                           feed_dict={"v": 5.0})

    def test_compile_const(self):
        a, b, w = Qbit("a"), Qbit("b"), Placeholder("w")
        exp = Constraint(Constraint(w * (a + b - 1), label="const1")
                         + Constraint((a + b - 1) ** 2, label="const2"), label="const_all")
        expected_qubo = {('a', 'a'): 2.0, ('a', 'b'): 2.0, ('b', 'b'): 2.0}
        expected_offset = -2.0
        expected_structure = {'a': ('a',), 'b': ('b',)}
        self.compile_check(exp, expected_qubo, expected_offset, expected_structure,
                           feed_dict={"w": 3.0})

    def test_compile_vector(self):
        x = Array.create('x', 5, 'BINARY')
        a = Qbit('a')
        exp = x[1] * (x[0] + 2*a + 1)
        expected_qubo = {
            ('a', 'a'): 0.0,
            ('a', 'x[1]'): 2.0,
            ('x[0]', 'x[0]'): 0.0,
            ('x[0]', 'x[1]'): 1.0,
            ('x[1]', 'x[1]'): 1.0}
        expected_offset = 0.0
        expected_structure = {
             'a': ('a',),
             'x[0]': ('x', 0),
             'x[1]': ('x', 1)}
        self.compile_check(exp, expected_qubo, expected_offset, expected_structure)

    def test_compile_vector_spin(self):
        x = Array.create('x', 2, 'SPIN')
        exp = x[1] * x[0] + x[0]
        expected_qubo = {('x[1]', 'x[1]'): -2.0, ('x[0]', 'x[0]'): 0.0, ('x[0]', 'x[1]'): 4.0}
        expected_offset = 0.0
        expected_structure = {
             'x[0]': ('x', 0),
             'x[1]': ('x', 1)}
        self.compile_check(exp, expected_qubo, expected_offset, expected_structure)

    def test_compile_matrix(self):
        x = Array.create("x", shape=(2, 2), vartype="BINARY")
        a = Qbit("a")
        exp = x[0, 1] * (x[1, 1] + 2*a + 1)
        expected_qubo = {
          ('a', 'a'): 0.0,
          ('a', 'x[0][1]'): 2.0,
          ('x[0][1]', 'x[0][1]'): 1.0,
          ('x[0][1]', 'x[1][1]'): 1.0,
          ('x[1][1]', 'x[1][1]'): 0.0}
        expected_offset = 0.0
        expected_structure = {
             'a': ('a',),
             'x[0][1]': ('x', 0, 1),
             'x[1][1]': ('x', 1, 1)}
        self.compile_check(exp, expected_qubo, expected_offset, expected_structure)

    def test_compile_matrix_spin(self):
        x = Array.create("x", (2, 2), "SPIN")
        exp = x[1, 1] * x[0, 0] + x[0, 0]
        expected_qubo = {('x[1][1]', 'x[1][1]'): -2.0, ('x[0][0]', 'x[0][0]'): 0.0,
                         ('x[0][0]', 'x[1][1]'): 4.0}
        expected_offset = 0.0
        expected_structure = {
             'x[0][0]': ('x', 0, 0),
             'x[1][1]': ('x', 1, 1)}
        self.compile_check(exp, expected_qubo, expected_offset, expected_structure)

    def test_index_error_vector(self):
        x = Array.create("x", 2, "BINARY")
        self.assertRaises(IndexError, lambda: x[2])

        x = Array.create("x", (2, 1), "BINARY")
        self.assertRaises(IndexError, lambda: x[2, 2])

    def test_sum(self):
        x = Array.create('x', 2, "BINARY")
        exp = (Sum(0, 2, lambda i: x[i]) - 1) ** 2

        expected_qubo = {('x[0]', 'x[0]'): -1.0, ('x[1]', 'x[1]'): -1.0, ('x[0]', 'x[1]'): 2.0}
        expected_offset = 1.0
        expected_structure = {
             'x[0]': ('x', 0),
             'x[1]': ('x', 1)}
        self.compile_check(exp, expected_qubo, expected_offset, expected_structure)



# Copyright 2018 Recruit Communications Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
from pyqubo import Qbit, Spin, Array, Num

import numpy as np


class TestArray(unittest.TestCase):

    def test_array_from_list(self):
        array = Array([[Qbit('x0'), Qbit('x1')], [Qbit('x2'), Qbit('x3')]])
        self.assertTrue(array.shape == (2, 2))
        self.assertTrue(array[0, 0] == Qbit('x0'))
        self.assertTrue(array[:, 0] == Array([Qbit('x0'), Qbit('x2')]))
        self.assertTrue(array[[0, 1], 0] == Array([Qbit('x0'), Qbit('x2')]))
        self.assertRaises(TypeError, lambda: Array(1))
        self.assertTrue(Array([[1, 2], [3, 4]]) == Array([Array([1, 2]), Array([3, 4])]))
        self.assertTrue(Array([[1, 2], [3, 4]]) == Array([np.array([1, 2]), np.array([3, 4])]))

    def test_array_error(self):
        # raise ValueError when the size of the sub lists is not same
        with self.assertRaises(ValueError):
            Array([[Qbit('x0'), Qbit('x1')], [Qbit('x2')]])

        # raise TypeError when index is neither int nor tuple[int]
        with self.assertRaises(TypeError):
            array = Array([Qbit('x0'), Qbit('x1')])
            array[1.5]

    def test_array_equality(self):
        array1 = Array([[Qbit('x0'), Qbit('x1')], [Qbit('x2'), Qbit('x3')]])
        array2 = Array([[Qbit('x0'), Qbit('x1')], [Qbit('x2'), Qbit('x3')]])
        array3 = Array([[Qbit('x0'), Qbit('x1')], [Qbit('x2'), Qbit('x4')]])
        self.assertTrue(array1 == array2)
        self.assertTrue(array1 != array3)
        self.assertTrue(not array1 == Qbit('x2'))

    def test_array_create_spin(self):
        array = Array.create('x', shape=(3, 3, 3), vartype='SPIN')
        self.assertTrue(array.shape == (3, 3, 3))
        self.assertTrue(array[0][0][0] == Spin('x[0][0][0]'))

    def test_array_create_binary(self):
        array = Array.create('x', shape=(3, 3, 3), vartype='BINARY')
        self.assertTrue(array.shape == (3, 3, 3))
        self.assertTrue(array[0][0][0] == Qbit('x[0][0][0]'))

    def test_array_repr(self):
        array = Array.create('x', shape=(3, 3, 3), vartype='BINARY')
        expected_string =\
            'Array([[[Qbit(x[0][0][0]), Qbit(x[0][0][1]), Qbit(x[0][0][2])],\n'\
            '        [Qbit(x[0][1][0]), Qbit(x[0][1][1]), Qbit(x[0][1][2])],\n'\
            '        [Qbit(x[0][2][0]), Qbit(x[0][2][1]), Qbit(x[0][2][2])]],\n'\
            '\n'\
            '       [[Qbit(x[1][0][0]), Qbit(x[1][0][1]), Qbit(x[1][0][2])],\n'\
            '        [Qbit(x[1][1][0]), Qbit(x[1][1][1]), Qbit(x[1][1][2])],\n'\
            '        [Qbit(x[1][2][0]), Qbit(x[1][2][1]), Qbit(x[1][2][2])]],\n'\
            '\n'\
            '       [[Qbit(x[2][0][0]), Qbit(x[2][0][1]), Qbit(x[2][0][2])],\n'\
            '        [Qbit(x[2][1][0]), Qbit(x[2][1][1]), Qbit(x[2][1][2])],\n'\
            '        [Qbit(x[2][2][0]), Qbit(x[2][2][1]), Qbit(x[2][2][2])]]])'
        self.assertTrue(repr(array) == expected_string)

    def test_array_iter(self):
        array = Array.create('x', shape=(2, 2), vartype='BINARY')
        self.assertTrue(len(array) == 2)
        self.assertTrue([e for e in array] == [array[0], array[1]])

    def test_array_from_numpy(self):
        array = Array(np.array([[1, 2], [3, 4]]))
        self.assertTrue(array == Array([[1, 2], [3, 4]]))

    def test_array_transpose(self):
        array = Array.create('x', (2, 3), 'BINARY')
        expected = Array([[Qbit('x[0][0]'), Qbit('x[1][0]')],
                          [Qbit('x[0][1]'), Qbit('x[1][1]')],
                          [Qbit('x[0][2]'), Qbit('x[1][2]')]])
        self.assertTrue(array.T == expected)

    def test_array_add(self):
        array1 = Array.create('x', (2, 2), 'BINARY')
        array2 = Array.create('y', (2, 2), 'BINARY')
        expected = Array([[Qbit('x[0][0]') + Qbit('y[0][0]'), Qbit('x[0][1]') + Qbit('y[0][1]')],
                          [Qbit('x[1][0]') + Qbit('y[1][0]'), Qbit('x[1][1]') + Qbit('y[1][1]')]])
        self.assertTrue(array1 + array2 == expected)

        expected2 = Array([[Qbit('x[0][0]') + 1, Qbit('x[0][1]') + 1],
                          [Qbit('x[1][0]') + 1, Qbit('x[1][1]') + 1]])
        self.assertTrue(array1 + 1 == expected2)
        self.assertTrue(1 + array1 == expected2)
        self.assertTrue(array1 + np.ones((2, 2)) == expected2)

        self.assertRaises(TypeError, lambda: array1 + "str")
        array3 = Array.create('z', (2, 3), 'BINARY')
        self.assertRaises(ValueError, lambda: array1 + array3)

    def test_array_subtract(self):
        array1 = Array.create('x', (2, 2), 'BINARY')
        array2 = Array.create('y', (2, 2), 'BINARY')
        expected = Array([[Qbit('x[0][0]') - Qbit('y[0][0]'), Qbit('x[0][1]') - Qbit('y[0][1]')],
                          [Qbit('x[1][0]') - Qbit('y[1][0]'), Qbit('x[1][1]') - Qbit('y[1][1]')]])
        self.assertTrue(array1 - array2 == expected)

        expected2 = Array([[Qbit('x[0][0]') - 1, Qbit('x[0][1]') - 1],
                          [Qbit('x[1][0]') - 1, Qbit('x[1][1]') - 1]])
        self.assertTrue(array1 - 1 == expected2)
        expected3 = Array([[1 - Qbit('x[0][0]'), 1 - Qbit('x[0][1]')],
                           [1 - Qbit('x[1][0]'), 1 - Qbit('x[1][1]')]])
        self.assertTrue(1 - array1 == expected3)
        self.assertTrue(array1 - np.ones((2, 2)) == expected2)

    def test_array_multiply(self):
        array1 = Array.create('x', (2, 2), 'BINARY')
        array2 = Array.create('y', (2, 2), 'BINARY')
        expected = Array([[Qbit('x[0][0]') * Qbit('y[0][0]'), Qbit('x[0][1]') * Qbit('y[0][1]')],
                          [Qbit('x[1][0]') * Qbit('y[1][0]'), Qbit('x[1][1]') * Qbit('y[1][1]')]])
        self.assertTrue(array1 * array2 == expected)

        expected2 = Array([[Qbit('x[0][0]') * 2, Qbit('x[0][1]') * 2],
                           [Qbit('x[1][0]') * 2, Qbit('x[1][1]') * 2]])
        self.assertTrue(array1 * 2 == expected2)
        self.assertTrue(2 * array1 == expected2)
        self.assertTrue(array1 * (2*np.ones((2, 2))) == expected2)

    def test_array_division(self):
        array1 = Array.create('x', (2, 2), 'BINARY')
        expected = Array([[Qbit('x[0][0]') / 2.0, Qbit('x[0][1]') / 2.0],
                           [Qbit('x[1][0]') / 2.0, Qbit('x[1][1]') / 2.0]])
        self.assertTrue(array1 / 2.0 == expected)

        self.assertRaises(ValueError, lambda: 1 / array1)
        self.assertRaises(ValueError, lambda: array1 / array1)

    def test_array_dot(self):

        # inner product of 1-D arrays
        array_a = Array([Qbit('a'), Qbit('b')])
        array_b = Array([Qbit('c'), Qbit('d')])
        expected = ((Qbit('a') * Qbit('c')) + (Qbit('b') * Qbit('d')))
        self.assertTrue(expected == array_a.dot(array_b))

        # dot product of 1-D array and N-D array
        array_a = Array([[Qbit('a'), Qbit('b')], [Qbit('c'), Qbit('d')]])
        array_b = Array([Qbit('e'), Qbit('f')])
        expected = Array([((Qbit('a') * Qbit('e')) + (Qbit('b') * Qbit('f'))),
                          ((Qbit('c') * Qbit('e')) + (Qbit('d') * Qbit('f')))])
        self.assertTrue(expected == array_a.dot(array_b))

        # both are 2-D arrays
        array_a = Array([[Qbit('a'), Qbit('b')], [Qbit('c'), Qbit('d')]])
        array_b = Array([[Qbit('e'), Qbit('f')], [Qbit('g'), Qbit('h')]])
        expected = Array([[((Qbit('a') * Qbit('e')) + (Qbit('b') * Qbit('g'))),
                           ((Qbit('a') * Qbit('f')) + (Qbit('b') * Qbit('h')))],
                          [((Qbit('c') * Qbit('e')) + (Qbit('d') * Qbit('g'))),
                           ((Qbit('c') * Qbit('f')) + (Qbit('d') * Qbit('h')))]])
        self.assertTrue(expected == array_a.dot(array_b))

        # array_a is a  N-D array and array_b is an M-D array (where N, M>=2)
        array_a = Array.create('a', shape=(3, 2, 4), vartype='BINARY')
        array_b = Array.create('b', shape=(5, 4, 3), vartype='BINARY')
        i, j, k, m = (1, 1, 3, 2)
        self.assertTrue(array_a.dot(array_b)[i, j, k, m]
                        == sum(array_a[i, j, :] * array_b[k, :, m]))

        # array_a is 1-D array and array_b is a list
        array_a = Array([Qbit('a'), Qbit('b')])
        array_b = [3, 4]
        self.assertTrue((Qbit('a')*3 + (Qbit('b')*4)) == array_a.dot(array_b))

        # test validation
        self.assertRaises(TypeError, lambda: array_a.dot(1))

    def test_array_matmul(self):

        # the either of the arguments is 1-D array,
        array_a = Array([[Qbit('a'), Qbit('b')], [Qbit('c'), Qbit('d')]])
        array_b = Array([Qbit('e'), Qbit('f')])
        expected = Array([((Qbit('a')*Qbit('e'))+(Qbit('b')*Qbit('f'))),
                          ((Qbit('c')*Qbit('e'))+(Qbit('d')*Qbit('f')))])
        self.assertTrue(array_a.matmul(array_b) == expected)

        # both arguments are 2-D array
        array_a = Array([[Qbit('a'), Qbit('b')], [Qbit('c'), Qbit('d')]])
        array_b = Array([[Qbit('e'), Qbit('f')], [Qbit('g'), Qbit('h')]])
        expected = Array([[((Qbit('a') * Qbit('e')) + (Qbit('b') * Qbit('g'))),
                           ((Qbit('a') * Qbit('f')) + (Qbit('b') * Qbit('h')))],
                          [((Qbit('c') * Qbit('e')) + (Qbit('d') * Qbit('g'))),
                           ((Qbit('c') * Qbit('f')) + (Qbit('d') * Qbit('h')))]])
        self.assertTrue(array_a.matmul(array_b) == expected)

        # either argument is N-D (where N > 2)
        array_a = Array.create('a', shape=(2, 2, 3), vartype='BINARY')
        array_b = Array.create('b', shape=(3, 2), vartype='BINARY')
        self.assertTrue((array_a.matmul(array_b))[0] == array_a[0].matmul(array_b))

        # array_a is 2-D array and array_b is a list
        array_a = Array([[Qbit('a'), Qbit('b')], [Qbit('c'), Qbit('d')]])
        array_b = [3, 4]
        expected = Array([((Qbit('a')*Num(3))+(Qbit('b')*Num(4))), ((Qbit('c')*Num(3))+(Qbit('d')*Num(4)))])
        self.assertTrue(expected == array_a.matmul(array_b))

        # test validation
        array_a = Array.create('a', shape=(2, 2, 3), vartype='BINARY')
        array_b = Array.create('b', shape=(3, 3, 2), vartype='BINARY')
        self.assertRaises(AssertionError, lambda: array_a.matmul(array_b))

    def test_array_reshape(self):
        array = Array.create('a', shape=(2, 3), vartype='BINARY')
        reshaped = array.reshape((3, 2))
        self.assertTrue(reshaped.shape == (3, 2))
        expected = Array([[Qbit('a[0][0]'), Qbit('a[0][1]')],
                          [Qbit('a[0][2]'), Qbit('a[1][0]')],
                          [Qbit('a[1][1]'), Qbit('a[1][2]')]])
        self.assertTrue(reshaped == expected)



# Copyright 2018 Recruit Communications Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import neal
import numpy as np


def solve_qubo(qubo, num_reads=10, sweeps=1000, beta_range=(1.0, 50.0)):
    """Solve QUBO with Simulated Annealing (SA) provided by neal.
    
    Args:
        qubo (dict[(label, label), float]): The QUBO to be solved.
        
        num_reads (int, default=10): Number of run repetitions of SA.
        
        sweeps (int, default=1000): Number of iterations in each run of SA.
         
        beta_range (tuple(float, float), default=(1.0, 50.0)): Tuple of start beta and end beta.
         
    Returns:
         dict[label, bit]: The solution of SA.
    
    >>> from pyqubo import Spin, solve_qubo
    >>> s1, s2, s3 = Spin("s1"), Spin("s2"), Spin("s3")
    >>> H = (2*s1 + 4*s2 + 6*s3)**2
    >>> model = H.compile()
    >>> qubo, offset = model.to_qubo()
    >>> solution = solve_qubo(qubo)
    """
    max_abs_value = float(max(abs(v) for v in qubo.values()))
    scale_qubo = {k: float(v) / max_abs_value for k, v in qubo.items()}
    sa = neal.SimulatedAnnealingSampler()
    sa_computation = sa.sample_qubo(scale_qubo, num_reads=num_reads,
                                    sweeps=sweeps, beta_range=beta_range)
    best = np.argmin(sa_computation.record.energy)
    best_solution = list(sa_computation.record.sample[best])
    return dict(zip(sa_computation.variable_labels, best_solution))


def solve_ising(linear, quad, num_reads=10, sweeps=1000, beta_range=(1.0, 50.0)):
    """Solve Ising model with Simulated Annealing (SA) provided by neal.

    Args:
        linear (dict[label, float]): The linear parameter of the Ising model.
        
        quad (dict[(label, label), float]): The quadratic parameter of the Ising model.

        num_reads (int, default=10): Number of run repetitions of SA.

        sweeps (int, default=1000): Number of iterations in each run of SA.

        beta_range (tuple(float, float), default=(1.0, 50.0)): Tuple of start beta and end beta.

    Returns:
         dict[label, bit]: The solution of SA.
    
    >>> from pyqubo import Spin, solve_ising
    >>> s1, s2, s3 = Spin("s1"), Spin("s2"), Spin("s3")
    >>> H = (2*s1 + 4*s2 + 6*s3)**2
    >>> model = H.compile()
    >>> linear, quad, offset = model.to_ising()
    >>> solution = solve_ising(linear, quad)
    """
    max_abs_value = float(max(abs(v) for v in (list(quad.values()) + list(linear.values()))))
    scale_linear = {k: float(v) / max_abs_value for k, v in linear.items()}
    scale_quad = {k: float(v) / max_abs_value for k, v in quad.items()}
    sa = neal.SimulatedAnnealingSampler()
    sa_computation = sa.sample_ising(scale_linear, scale_quad, num_reads=num_reads,
                                     sweeps=sweeps, beta_range=beta_range)
    best = np.argmin(sa_computation.record.energy)
    best_solution = list(sa_computation.record.sample[best])
    return dict(zip(sa_computation.variable_labels, best_solution))



# Copyright 2018 Recruit Communications Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
from pyqubo import assert_qubo_equal


class TestAsserts(unittest.TestCase):

    def test_qubo_equal(self):
        qubo1 = {('a', 'b'): 1.0}
        qubo2 = {('b', 'a'): 1.0}
        qubo3 = {('a', 'b'): 2.0}
        qubo4 = {('b', 'a'): 2.0}
        qubo5 = {('c', 'a'): 1.0}

        assert_qubo_equal(qubo1, qubo2)
        self.assertRaises(AssertionError, lambda: assert_qubo_equal(qubo1, qubo3))
        self.assertRaises(AssertionError, lambda: assert_qubo_equal(qubo1, qubo4))
        self.assertRaises(AssertionError, lambda: assert_qubo_equal(qubo1, qubo5))



from __future__ import absolute_import
from setuptools import setup
import os


class PackageInfo(object):
    def __init__(self, info_file):
        with open(info_file) as f:
            exec(f.read(), self.__dict__)
        self.__dict__.pop('__builtins__', None)

    def __getattribute__(self, name):
        return super(PackageInfo, self).__getattribute__(name)

package_info = PackageInfo(os.path.join('pyqubo', 'package_info.py'))


install_requires = ['dimod>=0.7.4',
                    'numpy>=1.14.0,<2.0.0',
                    'six>=1.10.0,<2.0.0',
                    'dwave-neal>=0.4.2']

packages = ['pyqubo', 'pyqubo.core', 'pyqubo.utils']

python_requires = '>=2.7,!=3.0.*,!=3.1.*,!=3.2.*,!=3.3.*'

setup(
    name=package_info.__package_name__,
    version=package_info.__version__,
    description=package_info.__description__,
    long_description=open('README.rst').read(),
    author=package_info.__contact_names__,
    author_email=package_info.__contact_emails__,
    maintainer=package_info.__contact_names__,
    maintainer_email=package_info.__contact_emails__,
    url=package_info.__repository_url__,
    download_url=package_info.__download_url__,
    license=package_info.__license__,
    packages=packages,
    keywords=package_info.__keywords__,
    install_requires=install_requires,
    include_package_data=True,
    python_requires=python_requires,
    classifiers=[
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'License :: OSI Approved :: Apache Software License',
    ],
)



# Copyright 2018 Recruit Communications Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .core import Spin, Qbit, Express

import dimod
from dimod.decorators import vartype_argument
import numpy as np
from operator import mul, add
from six.moves import reduce


class Array:
    """Multi-dimensional array.

    Args:
        bit_list (list/:class:`numpy.ndarray`): The object from which a new array is created.
            Accepted input:
        
                * (Nested) list of :class:`Express`, :class:`Array`, int or float.
                * numpy.ndarray

    Attributes:
        shape (tuple[int]): Shape of this array.

    Example:
        
        Create a new array with Qbits.
        
        >>> from pyqubo import Array, Qbit
        >>> Array.create('x', shape=(2, 2), vartype='BINARY')
        Array([[Qbit(x[0][0]), Qbit(x[0][1])],
               [Qbit(x[1][0]), Qbit(x[1][1])]])
        
        Create a new array from a nested list of :class:`Express`.
        
        >>> array = Array([[Qbit('x0'), Qbit('x1')], [Qbit('x2'), Qbit('x3')]])
        >>> array
        Array([[Qbit(x0), Qbit(x1)],
               [Qbit(x2), Qbit(x3)]])
        
        Get the shape of the array.
        
        >>> array.shape
        (2, 2)
        
        Access an element with index.
        
        >>> array[0, 0]  # = array[(0, 0)]
        Qbit(x0)
        
        Use slice ":" to select a subset of the array.
        
        >>> array[:, 1]  # = array[(slice(None), 1)]
        Array([Qbit(x1), Qbit(x3)])
        >>> sum(array[:, 1])
        (Qbit(x1)+Qbit(x3))
        
        Use list or tuple to select a subset of the array.

        >>> array[[0, 1], 1]
        Array([Qbit(x1), Qbit(x3)])
        >>> array[(0, 1), 1]
        Array([Qbit(x1), Qbit(x3)])
        
        Create an array from numpy array.
        
        >>> import numpy as np
        >>> Array(np.array([[1, 2], [3, 4]]))
        Array([[1, 2],
               [3, 4]])
        
        Create an array from list of :class:`Array`.
        
        >>> Array([Array([1, 2]), Array([3, 4])])
        Array([[1, 2],
               [3, 4]])
    """

    def __init__(self, bit_list):

        if isinstance(bit_list, np.ndarray):
            self.shape = bit_list.shape
            self.bit_list = bit_list.tolist()

        elif isinstance(bit_list, list):
            def get_shape(l):
                if isinstance(l, list) or isinstance(l, Array) or isinstance(l, np.ndarray):
                    length = len(l)
                    shape_set = {get_shape(e) for e in l}
                    if len(shape_set) == 1:
                        sub_shape = shape_set.pop()
                        return tuple([length] + list(sub_shape))
                    else:
                        raise ValueError('Cannot determine the shape of input nested list.')
                else:
                    return tuple()

            def normalize_type(l):
                if isinstance(l, list):
                    return [normalize_type(e) for e in l]
                elif isinstance(l, Array):
                    return [normalize_type(e) for e in l.bit_list]
                elif isinstance(l, np.ndarray):
                    return [normalize_type(e) for e in l.tolist()]
                else:
                    return l

            self.shape = get_shape(bit_list)
            self.bit_list = normalize_type(bit_list)

        else:
            raise TypeError('argument should be ndarray or list')

    def __len__(self):
        return self.shape[0]

    def __getitem__(self, key):
        """Get a subset of this array.
        
        Args:
            key (int/tuple[int]): Index of array.
        
        Returns:
            :class:`Express`/:class:`Array`/int/float
        
        Example:
            >>> array = Array.create('x', (2, 3, 2), 'BINARY')
            >>> array
            Array([[[Qbit(x[0][0][0]), Qbit(x[0][0][1])],
                    [Qbit(x[0][1][0]), Qbit(x[0][1][1])],
                    [Qbit(x[0][2][0]), Qbit(x[0][2][1])]],
            
                   [[Qbit(x[1][0][0]), Qbit(x[1][0][1])],
                    [Qbit(x[1][1][0]), Qbit(x[1][1][1])],
                    [Qbit(x[1][2][0]), Qbit(x[1][2][1])]]])
            >>> array[0, 1, 1]
            Qbit(x[0][1][1])
            >>> array[:, :, 1]
            
        """
        if isinstance(key, int):
            key = key,
        elif not isinstance(key, tuple):
            raise TypeError("Key should be int or tuple of int")

        def get_item(l, index):
            if len(index) > 1:
                current_index = index[0]
                if isinstance(current_index, int):
                    return get_item(l[current_index], index[1:])
                elif isinstance(current_index, list) or isinstance(current_index, tuple):
                    return [get_item(l[i], index[1:]) for i in current_index]
                else:
                    return [get_item(e, index[1:]) for e in l[current_index]]
            else:
                return l[index[0]]

        item = get_item(self.bit_list, key)

        if isinstance(item, list):
            return Array(item)
        else:
            return item

    def __repr__(self):
        nest_depth = len(self.shape)
        offset = len("Array(")

        def format_nested_list(nested_list, nest_count):
            if isinstance(nested_list[0], list):
                return '[{body}]'.format(
                    body=',{line_feed}{indent}'.format(
                        indent=' ' * (nest_count + offset),
                        line_feed='\n' * (nest_depth - nest_count)
                    ).join([format_nested_list(sub_list, nest_count+1) for sub_list in nested_list])
                )
            else:
                return '[%s]' % ', '.join(map(str, nested_list))

        return 'Array({body})'.format(body=format_nested_list(self.bit_list, 1))

    def __eq__(self, other):
        if not isinstance(other, Array):
            return False
        else:
            return self.bit_list == other.bit_list

    def __ne__(self, other):
        return not self.__eq__(other)

    # math operation
    def __neg__(self):
        minus_one = Array.fill(-1, self.shape)
        return self * minus_one

    def __radd__(self, other):
        """It is called when `other(number) + self`"""
        return self.__add__(other)

    def __add__(self, other):
        """It is called when `self + other(any object)`"""
        return self.add(other)

    def __rsub__(self, other):
        """It is called when `other(number) - self`"""
        return (-self).add(other)

    def __sub__(self, other):
        """It is called when `self - other(any object)`"""
        return self.subtract(other)

    def __rmul__(self, other):
        """It is called when `other(number) * self`"""
        return self.__mul__(other)

    def __mul__(self, other):
        """It is called when `self * other(any object)`"""
        return self.mul(other)

    def __div__(self, other):
        """It is called when `self / other(any object)`"""
        return self.div(other)

    def __rdiv__(self, other):
        """It is called when `other(number) / self`"""
        raise ValueError("Number cannot be divided by Expression.")

    def __truediv__(self, other):  # pragma: no cover
        """division in Python3"""
        return self.__div__(other)

    def __rtruediv__(self, other):  # pragma: no cover
        """It is called when `other(number) / self`"""
        return self.__rdiv__(other)

    def __matmul__(self, other):  # pragma: no cover
        return self.matmul(other)

    def add(self, other):
        """Returns a sum of self and other.
        
        Args:
            other (:class:`Array`/:class:`ndarray`/int/float): Addend.
        
        Returns:
            :class:`Array`
        
        Example:
            
            >>> from pyqubo import Array, Qbit
            >>> import numpy as np
            >>> array_a = Array([[Qbit('a'), Qbit('b')], [Qbit('c'), 2]])
            >>> array_b = Array([[Qbit('d'), 1], [Qbit('f'), Qbit('g')]])
            >>> array_a.add(array_b)
            Array([[(Qbit(a)+Qbit(d)), (Qbit(b)+Num(1))],
                   [(Qbit(c)+Qbit(f)), (Qbit(g)+Num(2))]])
            >>> array_a + array_b
            Array([[(Qbit(a)+Qbit(d)), (Qbit(b)+Num(1))],
                   [(Qbit(c)+Qbit(f)), (Qbit(g)+Num(2))]])

            Sum of self and scalar value.
            
            >>> array_a + 5
            Array([[(Qbit(a)+Num(5)), (Qbit(b)+Num(5))],
                   [(Qbit(c)+Num(5)), 7]])
            
            Sum of self and numpy ndarray.
            
            >>> array_a + np.array([[1, 2], [3, 4]])
            Array([[(Qbit(a)+Num(1)), (Qbit(b)+Num(2))],
                   [(Qbit(c)+Num(3)), 6]])
        """
        return self._pairwise_op_with_type_check(other, lambda x, y: x + y)

    def subtract(self, other):
        """Returns a difference between other and self.

        Args:
            other (:class:`Array`/:class:`ndarray`/int/float): Subtrahend.

        Returns:
            :class:`Array`

        Example:

            >>> from pyqubo import Array, Qbit
            >>> import numpy as np
            >>> array_a = Array([[Qbit('a'), Qbit('b')], [Qbit('c'), 2]])
            >>> array_b = Array([[Qbit('d'), 1], [Qbit('f'), Qbit('g')]])
            >>> array_a.subtract(array_b)
            Array([[(Qbit(a)+(Qbit(d)*Num(-1))), (Qbit(b)+Num(-1))],
                   [(Qbit(c)+(Qbit(f)*Num(-1))), ((Qbit(g)*Num(-1))+Num(2))]])
            >>> array_a - array_b
            Array([[(Qbit(a)+(Qbit(d)*Num(-1))), (Qbit(b)+Num(-1))],
                   [(Qbit(c)+(Qbit(f)*Num(-1))), ((Qbit(g)*Num(-1))+Num(2))]])

            Difference of self and scalar value.
            
            >>> array_a - 5
            Array([[(Qbit(a)+Num(-5)), (Qbit(b)+Num(-5))],
                   [(Qbit(c)+Num(-5)), -3]])
            
            Difference of self and numpy ndarray.
            
            >>> array_a - np.array([[1, 2], [3, 4]])
            Array([[(Qbit(a)+Num(-1)), (Qbit(b)+Num(-2))],
                   [(Qbit(c)+Num(-3)), -2]])
        """
        return self._pairwise_op_with_type_check(other, lambda x, y: x - y)

    def mul(self, other):
        """Returns a multiplicity of self by other.

        Args:
            other (:class:`Array`/:class:`ndarray`/int/float): Factor.

        Returns:
            :class:`Array`

        Example:

            >>> from pyqubo import Array, Qbit
            >>> import numpy as np
            >>> array_a = Array([[Qbit('a'), Qbit('b')], [Qbit('c'), 2]])
            >>> array_b = Array([[Qbit('d'), 1], [Qbit('f'), Qbit('g')]])
            >>> array_a.mul(array_b)
            Array([[(Qbit(a)*Qbit(d)), (Qbit(b)*Num(1))],
                   [(Qbit(c)*Qbit(f)), (Qbit(g)*Num(2))]])
            >>> array_a * array_b
            Array([[(Qbit(a)*Qbit(d)), (Qbit(b)*Num(1))],
                   [(Qbit(c)*Qbit(f)), (Qbit(g)*Num(2))]])

            Product of self and scalar value.

            >>> array_a * 5
            Array([[(Qbit(a)*Num(5)), (Qbit(b)*Num(5))],
                   [(Qbit(c)*Num(5)), 10]])

            Product of self and numpy ndarray.
            
            >>> array_a * np.array([[1, 2], [3, 4]])
            Array([[(Qbit(a)*Num(1)), (Qbit(b)*Num(2))],
                   [(Qbit(c)*Num(3)), 8]])
        """
        return self._pairwise_op_with_type_check(other, lambda x, y: x * y)

    def div(self, other):
        """Returns division of self by other.
        
        Args:
            other (int/float): Divisor.
        
        Returns:
            :class:`Array`
        
        Example:
            
            >>> from pyqubo import Array, Qbit
            >>> array_a = Array([[Qbit('a'), Qbit('b')], [Qbit('c'), 2]])
            >>> array_a / 5
            Array([[(Qbit(a)*Num(0.2)), (Qbit(b)*Num(0.2))],
                   [(Qbit(c)*Num(0.2)), 0.4]])
        """
        if not isinstance(other, Array):
            return self * (other ** -1)
        else:
            raise ValueError("Expression cannot be divided by Expression.")

    @staticmethod
    @vartype_argument('vartype')
    def create(name, shape, vartype):
        """Create a new array with Spins or Qbits.

        Args:
            name (str): Name of the matrix. It is used as a part of the label of variables.
                For example, if the name is 'x',
                the label of `(i, j)` th variable will be ``x[i][j]``.
            
            shape (int/tuple[int]): Dimensions of the array.
            
            vartype (:class:`dimod.Vartype`/str/set, optional):
                Variable type of the solution.
                Accepted input values:
                
                * :class:`.Vartype.SPIN`, ``'SPIN'``, ``{-1, 1}``
                * :class:`.Vartype.BINARY`, ``'BINARY'``, ``{0, 1}``

        Example:
            >>> from pyqubo import Array
            >>> array = Array.create('x', shape=(2, 2), vartype='BINARY')
            >>> array
            Array([[Qbit(x[0][0]), Qbit(x[0][1])],
                   [Qbit(x[1][0]), Qbit(x[1][1])]])
            >>> array[0]
            Array([Qbit(x[0][0]), Qbit(x[0][1])])
        """

        if isinstance(shape, int):
            shape = shape,

        if vartype == dimod.BINARY:
            var_class = Qbit
        else:
            var_class = Spin

        def var_name(_name, index):
            return "{name}{index_repr}".format(
                name=_name, index_repr=''.join(['[%d]' % i for i in index]))

        def create_structure(index):
            return {var_name(name, index): tuple([name] + index)}

        def generator(index):
            return var_class(var_name(name, index), create_structure(index))

        return Array._create_with_generator(shape, generator)

    @staticmethod
    def fill(obj, shape):
        """Create a new array with the given shape, all filled with the given object.
        
        Args:
            obj (int/float/:class:`Express`): The object with which a new array is filled. 
            shape (tuple[int]): Shape of the array. 
        
        Returns:
            :class:`Array`: Created array.
        
        Example:
            
            >>> from pyqubo import Array, Qbit
            >>> Array.fill(Qbit('a'), shape=(2, 3))
            Array([[Qbit(a), Qbit(a), Qbit(a)],
                   [Qbit(a), Qbit(a), Qbit(a)]])
        """
        return Array._create_with_generator(shape, lambda _: obj)

    @staticmethod
    def _create_with_generator(shape, generator):
        """Returns an array with objects which `generator` created.
        
        Args:
            shape (tuple[int]): Shape of the array.
            generator (list[int] =>:class:`Express`): Function to generate :class:`Express`:.
                Type of the argument of the generator is ``list[int]``.
        
        Returns:
            :class:`Array`: Created array.
        """

        _shape_list = list(shape)

        def create_internal(shape_list, index):
            if len(shape_list) > 1:
                length = shape_list[0]
                return [create_internal(shape_list[1:], index + [i]) for i in range(length)]
            else:
                length = shape_list[0]
                return [generator(index+[i]) for i in range(length)]

        return Array(create_internal(_shape_list, []))

    def _pairwise_op_with_type_check(self, other, operation):
        """Pairwise operation with type check.
        
        Args:
            other (:class:`Array`/:class:`ndarray`/int/float): The other object in operation.
            operation (:class:`Express`, :class:`Express` => :class:`Express`): Operation.
        
        Returns:
            :class:`Array`
        """
        if isinstance(other, np.ndarray):
            other = Array(other)
        elif isinstance(other, int) or isinstance(other, float) or isinstance(other, Express):
            other = Array.fill(other, self.shape)
        elif not isinstance(other, Array):
            raise TypeError('Operation of Array cannot be done with type:{type}'
                            .format(type=type(other)))
        return self._pairwise_op(other, operation)

    def _pairwise_op(self, other, operation):
        """Pairwise operation
        
        Args:
            other (:class:`Array`): The other object in operation.
            operation (:class:`Express`, :class:`Express` => :class:`Express`): Operation
        
        Returns:
            :class:`Array`
        """
        if not isinstance(other, Array):  # pragma: no cover
            raise TypeError('Type of `other` is not a `Array` instance.')
        elif not self.shape == other.shape:
            raise ValueError('Shape of other is not same as that of self.')
        else:
            def operate(l1, l2):
                if isinstance(l1, list):
                    return [operate(e1, e2) for e1, e2 in zip(l1, l2)]
                else:
                    return operation(l1, l2)
            return Array(operate(self.bit_list, other.bit_list))

    @property
    def T(self):
        """Returns a transposed array.
        
        Example:
            >>> from pyqubo import Array
            >>> array = Array.create('x', shape=(2, 3), vartype='BINARY')
            >>> array
            Array([[Qbit(x[0][0]), Qbit(x[0][1]), Qbit(x[0][2])],
                   [Qbit(x[1][0]), Qbit(x[1][1]), Qbit(x[1][2])]])
            >>> array.T
            Array([[Qbit(x[0][0]), Qbit(x[1][0])],
                   [Qbit(x[0][1]), Qbit(x[1][1])],
                   [Qbit(x[0][2]), Qbit(x[1][2])]])
        """

        def generator(index):
            return self[tuple(index[::-1])]

        return Array._create_with_generator(self.shape[::-1], generator)

    def dot(self, other):
        """Returns a dot product of two arrays.
        
        Args:
            other (:class:`Array`): Array.
        
        Returns:
            :class:`Express`/:class:`Array`
        
        Example:
            
            Dot calculation falls into four patterns.
            
            1. If both `self` and `other` are 1-D arrays, it is inner product of vectors.
            
            >>> from pyqubo import Array, Qbit
            >>> array_a = Array([Qbit('a'), Qbit('b')])
            >>> array_b = Array([Qbit('c'), Qbit('d')])
            >>> array_a.dot(array_b)
            ((Qbit(a)*Qbit(c))+(Qbit(b)*Qbit(d)))
            
            2. If `self` is an N-D array and `other` is a 1-D array,\
                it is a sum product over the last axis of `self` and `other`.
            
            >>> array_a = Array([[Qbit('a'), Qbit('b')], [Qbit('c'), Qbit('d')]])
            >>> array_b = Array([Qbit('e'), Qbit('f')])
            >>> array_a.dot(array_b)
            Array([((Qbit(a)*Qbit(e))+(Qbit(b)*Qbit(f))), ((Qbit(c)*Qbit(e))+(Qbit(d)*Qbit(f)))])
            
            3. If both `self` and `other` are 2-D arrays, it is matrix multiplication.
            
            >>> array_a = Array([[Qbit('a'), Qbit('b')], [Qbit('c'), Qbit('d')]])
            >>> array_b = Array([[Qbit('e'), Qbit('f')], [Qbit('g'), Qbit('h')]])
            >>> array_a.dot(array_b)
            Array([[((Qbit(a)*Qbit(e))+(Qbit(b)*Qbit(g))), ((Qbit(a)*Qbit(f))+(Qbit(b)*Qbit(h)))],
                   [((Qbit(c)*Qbit(e))+(Qbit(d)*Qbit(g))), ((Qbit(c)*Qbit(f))+(Qbit(d)*Qbit(h)))]])
            
            4. If `self` is an N-D array and `other` is an M-D array (where N, M>=2),\
                it is a sum product over the last axis of `self` and\
                the second-to-last axis of `other`. If N = M = 3,\
                (i, j, k, m) element of a dot product of `self` and `other` is:

            .. code-block:: python

                dot(self, other)[i,j,k,m] = sum(self[i,j,:] * other[k,:,m])

            >>> array_a = Array.create('a', shape=(3, 2, 4), vartype='BINARY')
            >>> array_a.shape
            (3, 2, 4)
            >>> array_b = Array.create('b', shape=(5, 4, 3), vartype='BINARY')
            >>> array_b.shape
            (5, 4, 3)
            >>> i, j, k, m = (1, 1, 3, 2)
            >>> array_a.dot(array_b)[i, j, k, m] == sum(array_a[i, j, :] * array_b[k, :, m])
            True
            
            Dot product with list.
            
            >>> array_a = Array([Qbit('a'), Qbit('b')])
            >>> array_b = [3, 4]
            >>> array_a.dot(array_b)
            ((Qbit(a)*Num(3))+(Qbit(b)*Num(4)))
        """
        if isinstance(other, np.ndarray) or isinstance(other, list):
            other = Array(other)

        if not isinstance(other, Array):
            raise TypeError("Type of argument should be Array")

        # pattern 1 (see docstring)
        if len(self.shape) == 1 and len(other.shape) == 1 and self.shape[0] == other.shape[0]:
            return sum(self.mul(other))

        # pattern 2
        elif len(self.shape) == 2 and len(other.shape) == 1:
            return Array([sum(v * other) for v in self])

        # pattern 3 and 4
        else:
            return self._dot_matrix(other)

    def _dot_matrix(self, other):
        """Returns a dot product of N-D array self and M-D array other (where N, M>=2).
        """
        assert isinstance(other, Array), "Type should be Array, not {type}".format(type=type(other))
        assert self.shape[-1] == other.shape[-2],\
            "self.shape[-1] should be equal other.shape[-2].\n" +\
            "For more details, see https://pyqubo.readthedocs.io/en/latest/reference/array.html"

        vector_indices = slice(0, self.shape[-1], None)
        new_shape = self.shape[:-1] + other.shape[:-2] + (other.shape[-1],)

        def generator(index):
            half = len(self.shape) - 1
            index_self = tuple(index[:half]) + (vector_indices,)
            index_other = tuple(index[half:-1]) + (vector_indices,) + (index[-1],)
            vector_self = self[index_self]
            vector_other = other[index_other]
            return sum(vector_self * vector_other)

        return Array._create_with_generator(new_shape, generator)

    def matmul(self, other):
        """Returns a matrix product of two arrays.
        
        Note:
            You can use operator symbol '@' instead of :obj:`matmul()`
            in Python 3.5 or later version.
        
        >>> from pyqubo import Array
        >>> array_a = Array.create('a', shape=(2, 4), vartype='BINARY')
        >>> array_b = Array.create('b', shape=(4, 3), vartype='BINARY')
        >>> array_a @ array_b == array_a.matmul(array_b)
        True
        
        Args:
            other (:class:`Array`/:class:`numpy.ndarray`/list): 
        
        Returns:
            :class:`Array`/:class:`Express`
        
        Example:
        
            Matrix product of two arrays falls into 3 patterns.
            
            1. If either of the arguments is 1-D array,
            it is treated as a matrix where one is added to its dimension.
            
            >>> from pyqubo import Array, Qbit
            >>> array_a = Array([[Qbit('a'), Qbit('b')], [Qbit('c'), Qbit('d')]])
            >>> array_b = Array([Qbit('e'), Qbit('f')])
            >>> array_a.matmul(array_b)
            Array([((Qbit(a)*Qbit(e))+(Qbit(b)*Qbit(f))), ((Qbit(c)*Qbit(e))+(Qbit(d)*Qbit(f)))])
            
            2. If both arguments are 2-D array, conventional matrix product is calculated.
            
            >>> array_a = Array([[Qbit('a'), Qbit('b')], [Qbit('c'), Qbit('d')]])
            >>> array_b = Array([[Qbit('e'), Qbit('f')], [Qbit('g'), Qbit('h')]])
            >>> array_a.matmul(array_b)
            Array([[((Qbit(a)*Qbit(e))+(Qbit(b)*Qbit(g))), ((Qbit(a)*Qbit(f))+(Qbit(b)*Qbit(h)))],
                   [((Qbit(c)*Qbit(e))+(Qbit(d)*Qbit(g))), ((Qbit(c)*Qbit(f))+(Qbit(d)*Qbit(h)))]])
            
            3. If either argument is N-D (where N > 2), it is treated as an array whose element is a
            2-D matrix of last two indices. In this example, `array_a` is treated as if
            it is a vector whose elements are two matrices of shape (2, 3).

            >>> array_a = Array.create('a', shape=(2, 2, 3), vartype='BINARY')
            >>> array_b = Array.create('b', shape=(3, 2), vartype='BINARY')
            >>> (array_a @ array_b)[0] == array_a[0].matmul(array_b)
            True
        """

        if isinstance(other, np.ndarray) or isinstance(other, list):
            other = Array(other)
        assert isinstance(other, Array), "Type should be Array, not {type}".format(type=type(other))

        # pattern 1 (see docstring)
        if len(self.shape) == 1 or len(other.shape) == 1:
            return self.dot(other)

        # pattern 2 and 3
        else:
            return self._matmul_matrix(other)

    def _matmul_matrix(self, other):

        assert isinstance(other, Array), "Type should be Array, not {type}".format(type=type(other))
        assert len(self.shape) >= 2 and len(other.shape) >= 2, "Shape should be greater than 2"
        assert self.shape[-1] == other.shape[-2], \
            "self.shape[-1] should be equal other.shape[-2].\n" + \
            "For more details, see https://pyqubo.readthedocs.io/en/latest/reference/array.html"

        self_shape_len = len(self.shape)
        other_shape_len = len(other.shape)

        common_len = min(self_shape_len, other_shape_len)

        for s1, s2 in zip(self.shape[-common_len:-2], other.shape[-common_len:-2]):
            assert s1 == s2, "Shape doesn't match."

        longer_shape = self.shape if self_shape_len > other_shape_len else other.shape
        new_shape = longer_shape[:-2] + (self.shape[-2], other.shape[-1])

        def generator(index):
            mat_index_self = tuple(index[-self_shape_len:][:-2])
            mat_index_other = tuple(index[-other_shape_len:][:-2])
            mat_self = self[mat_index_self] if mat_index_self != () else self
            mat_other = other[mat_index_other] if mat_index_other != () else other
            j = index[-1]
            i = index[-2]
            return mat_self[i, :].dot(mat_other[:, j])

        return Array._create_with_generator(new_shape, generator)

    @staticmethod
    def _calc_steps(shape):
        """Returns steps of shape.
        
        Step is used to create an 1-dim index from n-dim index like
        
        >>> steps = Array._calc_steps(shape)
        >>> one_dim_index = sum(step * i for step, i in zip(steps, n_dim_index))
        """
        steps = []
        tmp_d = 1
        for d in shape[::-1]:
            steps.append(tmp_d)
            tmp_d *= d
        steps = steps[::-1]
        return steps

    def reshape(self, new_shape):
        """Returns a reshaped array.
        
        Args:
            new_shape (tuple[int]): New shape.
        
        Example:
            
            >>> from pyqubo import Array
            >>> array = Array.create('x', shape=(2, 3), vartype='BINARY')
            >>> array
            Array([[Qbit(x[0][0]), Qbit(x[0][1]), Qbit(x[0][2])],
                   [Qbit(x[1][0]), Qbit(x[1][1]), Qbit(x[1][2])]])
            >>> array.reshape((3, 2, 1))
            Array([[[Qbit(x[0][0])],
                    [Qbit(x[0][1])]],\
            
                   [[Qbit(x[0][2])],
                    [Qbit(x[1][0])]],\

                   [[Qbit(x[1][1])],
                    [Qbit(x[1][2])]]])

        """
        assert reduce(mul, self.shape) == reduce(mul, new_shape),\
            "cannot reshape array of size {p} into shape {new_shape}".format(
                p=reduce(mul, self.shape), new_shape=new_shape)

        def calc_one_dim_array(nested_list):
            if isinstance(nested_list, list):
                return reduce(add, [calc_one_dim_array(e) for e in nested_list])
            else:
                return [nested_list]

        # create an 1-dim array from the n-dim array
        one_dim_array = calc_one_dim_array(self.bit_list)

        new_steps = Array._calc_steps(new_shape)

        def generator(index):
            # create an index for 1-dim array from the given index
            one_dim_index = sum(step * i for step, i in zip(new_steps, index))
            return one_dim_array[one_dim_index]

        return Array._create_with_generator(new_shape, generator)



# Copyright 2018 Recruit Communications Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .core import UserDefinedExpress


class Sum(UserDefinedExpress):
    """Define sum of the expressions over sequent indices.
    
    Note:
        Indices run from :obj:`start_index` to :obj:`end_index-1`.

    Args:
        start_index (int): index to start with.
        
        end_index (int): index ends with end_index-1.
        
        func (function): function which takes integer as an argument and returns :class:`Express`.
    
    Example:
        >>> from pyqubo import Sum, Array
        >>> x = Array.create('x', 3, 'BINARY')
        >>> exp = (Sum(0, 3, lambda i: x[i]) - 1.0)**2
        >>> pprint(exp.compile().to_qubo())
        ({('x[0]', 'x[0]'): -1.0,
          ('x[0]', 'x[1]'): 2.0,
          ('x[0]', 'x[2]'): 2.0,
          ('x[1]', 'x[1]'): -1.0,
          ('x[1]', 'x[2]'): 2.0,
          ('x[2]', 'x[2]'): -1.0},
         1.0)
    """

    def __init__(self, start_index, end_index, func):
        express = sum(func(i) for i in range(start_index, end_index))
        super(Sum, self).__init__(express)



# Copyright 2018 Recruit Communications Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from operator import mul, xor
from six.moves import reduce
import copy


class PlaceholderProd(object):
    """A product of placeholder variables.
    This class is used as a key of dictionary when you represent a polynomial as a dictionary.

    For example, a polynomial :math:`2a^2 b + 2` is represented as
    
    .. code-block:: python
        
        {PlaceholderProd({'a': 2, 'b': 1}): 2.0, PlaceholderProd({}): 2.0}

    Note:
        PlaceholderProd initialized with empty key corresponds to constant.

    Args:
        keys (dict[label, int]):
            dictionary with a key being label, a int value being order of power.
    """

    JOINT_SYMBOL = "*"
    CONST_STRING = "const"

    def __init__(self, keys):
        assert isinstance(keys, dict)
        self.keys = keys
        if keys:
            self.cached_hash = reduce(xor, [hash(k) ^ hash(v) for k, v in self.keys.items()])
            self.is_const = False

        # When :obj:`keys` is empty dict, this object represents constant.
        else:
            self.cached_hash = 0
            self.is_const = True

    def __hash__(self):
        return self.cached_hash

    def __eq__(self, other):
        if not isinstance(other, PlaceholderProd):
            return False
        else:
            return self.keys == other.keys

    def __ne__(self, other):
        return not self.__eq__(other)

    def __repr__(self):
        if self.is_constant():
            return self.CONST_STRING
        else:
            return self.JOINT_SYMBOL.join(sorted(["{}^{}".format(k, v)
                                                  for k, v in self.keys.items()]))

    def is_constant(self):
        """Returns whether this is constant or not.

        Returns:
            bool
        """
        return self.is_const

    def calc_product(self, dict_values):
        """Returns the value of the product of binary variables.

        Args:
            dict_values (dict[label, float]): value of binary variable. 

        Returns:
            float
        """
        if self.is_constant():
            return 1.0
        else:
            return reduce(mul, [dict_values[k] ** v for k, v in self.keys.items()])

    @staticmethod
    def merge_term_key(term_key1, term_key2):
        term_key1_copy = copy.copy(term_key1.keys)
        for k, v in term_key2.keys.items():
            if k in term_key1_copy:
                term_key1_copy[k] += v
            else:
                term_key1_copy[k] = v
        return PlaceholderProd(term_key1_copy)



# Copyright 2018 Recruit Communications Co., Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from operator import mul, xor
from six.moves import reduce


class BinaryProd(object):
    """A product of binary variables.
    This class is used as a key of dictionary when you represent a polynomial as a dictionary.
    
    For example, a polynomial :math:`2ab + b + 2` is represented as
    
    .. code-block:: python
        
        {BinaryProd({'a', 'b'}): 2.0, BinaryProd({'b'}): 1.0, BinaryProd(set()): 2.0}
    
    This class represents product of binary variables.
    Since :math:`a=a^2=a^3=...` where :math:`a` is binary, a product of binary variables
    can be represented by a set of unique variables.
    For example, :math:`aabbc` can be simplified as :math:`abc`.
    For this reason, this class contains unique binary variables as a set.
    
    Note:
        BinaryProd initialized with empty key corresponds to constant.
    
    Args:
        keys (set[label]): set of variable labels.
    """

    JOINT_SYMBOL = "*"
    CONST_STRING = "const"

    def __init__(self, keys):
        assert isinstance(keys, set)
        self.keys = keys
        if keys:
            self.cached_hash = reduce(xor, [hash(key) for key in self.keys])
            self.string = self.JOINT_SYMBOL.join(sorted(self.keys))

        # When :obj:`keys` is empty set, this object represents constant.
        else:
            self.cached_hash = 0
            self.string = self.CONST_STRING

    def __hash__(self):
        return self.cached_hash

    def __eq__(self, other):
        if not isinstance(other, BinaryProd):
            return False
        else:
            return self.keys == other.keys

    def __ne__(self, other):
        return not self.__eq__(other)

    def __repr__(self):
        return self.string

    def is_constant(self):
        """Returns whether this is constant or not.
        
        Returns:
            bool
        """
        if self.string == self.CONST_STRING:
            return True
        else:
            return False

    def calc_product(self, dict_values):
        """Returns the value of the product of binary variables.
        
        Args:
            dict_values (dict[label, float]): value of binary variable. 
        
        Returns:
            float
        """
        if self.is_constant():
            return 1.0
        else:
            return reduce(mul, [dict_values[key] for key in self.keys])

    @staticmethod
    def merge_term_key(term_key1, term_key2):
        return BinaryProd(term_key1.keys | term_key2.keys)
